// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    expr::{Expr, ToExpr},
    int::{UInt, UIntType},
    intern::Interned,
    reset::Reset,
    source_location::SourceLocation,
    ty::{
        impl_match_values_as_self, CanonicalType, CanonicalTypeKind, CanonicalValue, Connect,
        DynCanonicalType, FixedType, Type, TypeEnum, Value, ValueEnum,
    },
    util::interned_bit,
};
use bitvec::slice::BitSlice;

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug, Default)]
pub struct ClockType;

impl ClockType {
    pub const fn new() -> Self {
        Self
    }
}

impl Type for ClockType {
    type Value = Clock;
    type CanonicalType = ClockType;
    type CanonicalValue = Clock;
    type MaskType = UIntType<1>;
    type MaskValue = UInt<1>;

    impl_match_values_as_self!();

    fn mask_type(&self) -> Self::MaskType {
        UIntType::new()
    }

    fn type_enum(&self) -> TypeEnum {
        TypeEnum::Clock(*self)
    }

    fn from_canonical_type(t: Self::CanonicalType) -> Self {
        t
    }

    fn canonical(&self) -> Self::CanonicalType {
        *self
    }

    fn source_location(&self) -> SourceLocation {
        SourceLocation::builtin()
    }

    fn as_dyn_canonical_type_impl(this: &Self) -> Option<&dyn DynCanonicalType> {
        Some(this)
    }
}

impl Connect<Self> for ClockType {}

impl CanonicalType for ClockType {
    const CANONICAL_TYPE_KIND: CanonicalTypeKind = CanonicalTypeKind::Clock;
}

impl FixedType for ClockType {
    fn fixed_type() -> Self {
        Self
    }
}

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug)]
pub struct Clock(pub bool);

impl ToExpr for Clock {
    type Type = ClockType;

    fn ty(&self) -> Self::Type {
        ClockType
    }

    fn to_expr(&self) -> Expr<Self> {
        Expr::from_value(self)
    }
}

impl Value for Clock {
    fn to_canonical(&self) -> <Self::Type as Type>::CanonicalValue {
        *self
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        interned_bit(this.0)
    }
}

impl CanonicalValue for Clock {
    fn value_enum_impl(this: &Self) -> ValueEnum {
        ValueEnum::Clock(*this)
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        interned_bit(this.0)
    }
}

#[derive(Copy, Clone, Debug, Eq, PartialEq, Hash, crate::Value)]
#[hdl(fixed_type, outline_generated)]
pub struct ClockDomain {
    pub clock: Clock,
    pub reset: Reset,
}

pub trait ToClock {
    fn to_clock(&self) -> Expr<Clock>;
}

impl<T: ?Sized + ToClock> ToClock for &'_ T {
    fn to_clock(&self) -> Expr<Clock> {
        (**self).to_clock()
    }
}

impl<T: ?Sized + ToClock> ToClock for &'_ mut T {
    fn to_clock(&self) -> Expr<Clock> {
        (**self).to_clock()
    }
}

impl<T: ?Sized + ToClock> ToClock for Box<T> {
    fn to_clock(&self) -> Expr<Clock> {
        (**self).to_clock()
    }
}

impl ToClock for Expr<Clock> {
    fn to_clock(&self) -> Expr<Clock> {
        *self
    }
}

impl ToClock for Clock {
    fn to_clock(&self) -> Expr<Clock> {
        self.to_expr()
    }
}

impl ToClock for bool {
    fn to_clock(&self) -> Expr<Clock> {
        self.to_expr().to_clock()
    }
}

impl ToClock for UInt<1> {
    fn to_clock(&self) -> Expr<Clock> {
        self.to_expr().to_clock()
    }
}
