// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
#![allow(clippy::type_complexity)]
use crate::{
    bundle::{BundleValue, TypeHintTrait},
    expr::{ops::VariantAccess, Expr, ToExpr},
    int::{UInt, UIntType},
    intern::{Intern, Interned, MemoizeGeneric},
    module::{
        EnumMatchVariantAndInactiveScopeImpl, EnumMatchVariantsIterImpl, ModuleBuilder,
        NormalModule, Scope,
    },
    source_location::SourceLocation,
    ty::{
        CanonicalType, CanonicalTypeKind, CanonicalValue, Connect, DynCanonicalType,
        DynCanonicalValue, DynType, MatchVariantAndInactiveScope, Type, TypeEnum, Value, ValueEnum,
    },
};
use bitvec::{order::Lsb0, slice::BitSlice, vec::BitVec, view::BitView};
use hashbrown::HashMap;
use std::{
    borrow::Cow,
    fmt,
    hash::{Hash, Hasher},
    iter::FusedIterator,
    marker::PhantomData,
};

#[derive(Copy, Clone, Debug, Hash, PartialEq, Eq)]
pub struct VariantType<T> {
    pub name: Interned<str>,
    pub ty: Option<T>,
}

pub struct FmtDebugInEnum<'a, T>(&'a VariantType<T>);

impl<T: fmt::Debug> fmt::Debug for FmtDebugInEnum<'_, T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let VariantType { name, ref ty } = *self.0;
        if let Some(ty) = ty {
            write!(f, "{name}({ty:?})")
        } else {
            write!(f, "{name}")
        }
    }
}

impl<T: fmt::Debug> fmt::Display for FmtDebugInEnum<'_, T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        fmt::Debug::fmt(self, f)
    }
}

impl<T> VariantType<T> {
    pub fn map_opt_ty<U, F: FnOnce(Option<T>) -> Option<U>>(self, f: F) -> VariantType<U> {
        let Self { name, ty } = self;
        VariantType { name, ty: f(ty) }
    }
    pub fn map_ty<U, F: FnOnce(T) -> U>(self, f: F) -> VariantType<U> {
        let Self { name, ty } = self;
        VariantType {
            name,
            ty: ty.map(f),
        }
    }
    pub fn as_ref_ty(&self) -> VariantType<&T> {
        VariantType {
            name: self.name,
            ty: self.ty.as_ref(),
        }
    }
    pub fn fmt_debug_in_enum(&self) -> FmtDebugInEnum<T> {
        FmtDebugInEnum(self)
    }
}

impl<T: Type> VariantType<T> {
    pub fn canonical(&self) -> VariantType<T::CanonicalType> {
        self.as_ref_ty().map_ty(T::canonical)
    }
    pub fn to_dyn(&self) -> VariantType<Interned<dyn DynType>> {
        self.as_ref_ty().map_ty(T::to_dyn)
    }
    pub fn canonical_dyn(&self) -> VariantType<Interned<dyn DynCanonicalType>> {
        self.as_ref_ty().map_ty(T::canonical_dyn)
    }
}

impl VariantType<Interned<dyn DynCanonicalType>> {
    pub fn from_canonical_type_helper_has_value<T: Type>(self, expected_name: &str) -> T {
        assert_eq!(&*self.name, expected_name, "variant name doesn't match");
        let Some(ty) = self.ty else {
            panic!("variant {expected_name} has no value but a value is expected");
        };
        T::from_dyn_canonical_type(ty)
    }
    pub fn from_canonical_type_helper_no_value(self, expected_name: &str) {
        assert_eq!(&*self.name, expected_name, "variant name doesn't match");
        assert!(
            self.ty.is_none(),
            "variant {expected_name} has a value but is expected to have no value"
        );
    }
}

#[derive(Clone, Eq)]
struct DynEnumTypeImpl {
    variants: Interned<[VariantType<Interned<dyn DynCanonicalType>>]>,
    name_indexes: HashMap<Interned<str>, usize>,
    bit_width: usize,
    is_storable: bool,
    is_castable_from_bits: bool,
}

impl fmt::Debug for DynEnumTypeImpl {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "DynEnumType ")?;
        f.debug_set()
            .entries(
                self.variants
                    .iter()
                    .map(|variant| variant.fmt_debug_in_enum()),
            )
            .finish()
    }
}

impl PartialEq for DynEnumTypeImpl {
    fn eq(&self, other: &Self) -> bool {
        self.variants == other.variants
    }
}

impl Hash for DynEnumTypeImpl {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.variants.hash(state);
    }
}

#[derive(Copy, Clone, Hash, PartialEq, Eq)]
pub struct DynEnumType(Interned<DynEnumTypeImpl>);

impl fmt::Debug for DynEnumType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

fn discriminant_bit_width_impl(variant_count: usize) -> usize {
    variant_count
        .next_power_of_two()
        .checked_ilog2()
        .unwrap_or(0) as usize
}

impl DynEnumType {
    #[track_caller]
    pub fn new(variants: Interned<[VariantType<Interned<dyn DynCanonicalType>>]>) -> Self {
        assert!(!variants.is_empty(), "zero-variant enums aren't yet supported: https://github.com/chipsalliance/firrtl-spec/issues/208");
        let mut name_indexes = HashMap::with_capacity(variants.len());
        let mut body_bit_width = 0usize;
        let mut is_storable = true;
        let mut is_castable_from_bits = true;
        for (index, &VariantType { name, ty }) in variants.iter().enumerate() {
            if let Some(old_index) = name_indexes.insert(name, index) {
                panic!("duplicate variant name {name:?}: at both index {old_index} and {index}");
            }
            if let Some(ty) = ty {
                assert!(
                    ty.is_passive(),
                    "variant type must be a passive type: {ty:?}"
                );
                body_bit_width = body_bit_width.max(ty.bit_width());
                is_storable &= ty.is_storable();
                is_castable_from_bits &= ty.is_castable_from_bits();
            }
        }
        let bit_width = body_bit_width
            .checked_add(discriminant_bit_width_impl(variants.len()))
            .unwrap_or_else(|| panic!("enum is too big: bit-width overflowed"));
        Self(
            DynEnumTypeImpl {
                variants,
                name_indexes,
                bit_width,
                is_storable,
                is_castable_from_bits,
            }
            .intern_sized(),
        )
    }
    pub fn discriminant_bit_width(self) -> usize {
        discriminant_bit_width_impl(self.variants().len())
    }
    pub fn is_passive(self) -> bool {
        true
    }
    pub fn is_storable(self) -> bool {
        self.0.is_storable
    }
    pub fn is_castable_from_bits(self) -> bool {
        self.0.is_castable_from_bits
    }
    pub fn bit_width(self) -> usize {
        self.0.bit_width
    }
    pub fn name_indexes(&self) -> &HashMap<Interned<str>, usize> {
        &self.0.name_indexes
    }
}

#[derive(Debug, Clone, Hash, PartialEq, Eq)]
pub struct DynEnum {
    ty: DynEnumType,
    variant_index: usize,
    variant_value: Option<DynCanonicalValue>,
}

impl DynEnum {
    #[track_caller]
    pub fn new_by_index(
        ty: DynEnumType,
        variant_index: usize,
        variant_value: Option<DynCanonicalValue>,
    ) -> Self {
        let variant = ty.variants()[variant_index];
        assert_eq!(
            variant_value.as_ref().map(|v| v.ty()),
            variant.ty,
            "variant value doesn't match type"
        );
        Self {
            ty,
            variant_index,
            variant_value,
        }
    }
    #[track_caller]
    pub fn new_by_name(
        ty: DynEnumType,
        variant_name: Interned<str>,
        variant_value: Option<DynCanonicalValue>,
    ) -> Self {
        let variant_index = ty.name_indexes()[&variant_name];
        Self::new_by_index(ty, variant_index, variant_value)
    }
    pub fn variant_index(&self) -> usize {
        self.variant_index
    }
    pub fn variant_value(&self) -> &Option<DynCanonicalValue> {
        &self.variant_value
    }
    pub fn variant_with_type(&self) -> VariantType<Interned<dyn DynCanonicalType>> {
        self.ty.variants()[self.variant_index]
    }
    pub fn variant_name(&self) -> Interned<str> {
        self.variant_with_type().name
    }
    pub fn variant_type(&self) -> Option<Interned<dyn DynCanonicalType>> {
        self.variant_with_type().ty
    }
    pub fn variant_with_value(&self) -> VariantType<&DynCanonicalValue> {
        self.variant_with_type()
            .map_opt_ty(|_| self.variant_value.as_ref())
    }
}

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug)]
pub struct VariantsHint {
    pub known_variants: Interned<[VariantType<Interned<dyn TypeHintTrait>>]>,
    pub more_variants: bool,
}

impl VariantsHint {
    pub fn new(
        known_variants: impl IntoIterator<Item = VariantType<Interned<dyn TypeHintTrait>>>,
        more_variants: bool,
    ) -> Self {
        let known_variants = Intern::intern_owned(Vec::from_iter(known_variants));
        Self {
            known_variants,
            more_variants,
        }
    }
    pub fn check_variant(
        self,
        index: usize,
        variant: VariantType<&dyn DynType>,
    ) -> Result<(), String> {
        let Some(&known_variant) = self.known_variants.get(index) else {
            return if self.more_variants {
                Ok(())
            } else {
                Err(format!(
                    "too many variants: name={:?} index={index}",
                    variant.name
                ))
            };
        };
        let VariantType {
            name: known_name,
            ty: type_hint,
        } = known_variant;
        let VariantType { name, ty } = variant;
        if name != known_name {
            Err(format!(
                "wrong variant name {name:?}, expected {known_name:?}"
            ))
        } else {
            match (ty, type_hint) {
                (Some(ty), Some(type_hint)) => type_hint.matches(ty),
                (None, None) => Ok(()),
                (None, Some(_)) => Err(format!(
                    "expected variant {name:?} to have type, no type provided"
                )),
                (Some(_), None) => Err(format!(
                    "expected variant {name:?} to have no type, but a type was provided"
                )),
            }
        }
    }
}

pub trait EnumType:
    Type<
        CanonicalType = DynEnumType,
        CanonicalValue = DynEnum,
        MaskType = UIntType<1>,
        MaskValue = UInt<1>,
        MatchActiveScope = Scope,
        MatchVariantAndInactiveScope = EnumMatchVariantAndInactiveScope<Self>,
        MatchVariantsIter = EnumMatchVariantsIter<Self>,
    > + Connect<Self>
where
    Self::Value: EnumValue + ToExpr<Type = Self>,
{
    type Builder;
    fn match_activate_scope(
        v: Self::MatchVariantAndInactiveScope,
    ) -> (Self::MatchVariant, Self::MatchActiveScope);
    fn builder() -> Self::Builder;
    fn variants(&self) -> Interned<[VariantType<Interned<dyn DynCanonicalType>>]>;
    fn variants_hint() -> VariantsHint;
    #[allow(clippy::result_unit_err)]
    fn variant_to_bits<VariantValue: ToExpr + Eq + Hash + Send + Sync + 'static + Clone>(
        &self,
        variant_index: usize,
        variant_value: Option<&VariantValue>,
    ) -> Result<Interned<BitSlice>, ()> {
        #[derive(Hash, Eq, PartialEq)]
        struct VariantToBitsMemoize<E, V>(PhantomData<(E, V)>);
        impl<E, V> Clone for VariantToBitsMemoize<E, V> {
            fn clone(&self) -> Self {
                *self
            }
        }
        impl<E, V> Copy for VariantToBitsMemoize<E, V> {}
        impl<
                E: EnumType<Value: EnumValue<Type = E>>,
                V: ToExpr + Eq + Hash + Send + Sync + 'static + Clone,
            > MemoizeGeneric for VariantToBitsMemoize<E, V>
        {
            type InputRef<'a> = (&'a E, usize, Option<&'a V>);
            type InputOwned = (E, usize, Option<V>);
            type InputCow<'a> = (Cow<'a, E>, usize, Option<Cow<'a, V>>);
            type Output = Result<Interned<BitSlice>, ()>;

            fn input_borrow(input: &Self::InputOwned) -> Self::InputRef<'_> {
                (&input.0, input.1, input.2.as_ref())
            }
            fn input_eq(a: Self::InputRef<'_>, b: Self::InputRef<'_>) -> bool {
                a == b
            }
            fn input_cow_into_owned(input: Self::InputCow<'_>) -> Self::InputOwned {
                (input.0.into_owned(), input.1, input.2.map(Cow::into_owned))
            }
            fn input_cow_borrow<'a>(input: &'a Self::InputCow<'_>) -> Self::InputRef<'a> {
                (&input.0, input.1, input.2.as_deref())
            }
            fn input_cow_from_owned<'a>(input: Self::InputOwned) -> Self::InputCow<'a> {
                (Cow::Owned(input.0), input.1, input.2.map(Cow::Owned))
            }
            fn input_cow_from_ref(input: Self::InputRef<'_>) -> Self::InputCow<'_> {
                (Cow::Borrowed(input.0), input.1, input.2.map(Cow::Borrowed))
            }
            fn inner(self, input: Self::InputRef<'_>) -> Self::Output {
                let (ty, variant_index, variant_value) = input;
                let ty = ty.canonical();
                let mut bits = BitVec::with_capacity(ty.bit_width());
                bits.extend_from_bitslice(
                    &variant_index.view_bits::<Lsb0>()[..ty.discriminant_bit_width()],
                );
                if let Some(variant_value) = variant_value {
                    bits.extend_from_bitslice(&variant_value.to_expr().to_literal_bits()?);
                }
                bits.resize(ty.bit_width(), false);
                Ok(Intern::intern_owned(bits))
            }
        }
        VariantToBitsMemoize::<Self, VariantValue>(PhantomData).get((
            self,
            variant_index,
            variant_value,
        ))
    }
}

pub trait EnumValue: Value
where
    <Self as ToExpr>::Type: EnumType<Value = Self>,
{
}

pub struct EnumMatchVariantAndInactiveScope<T: EnumType>(EnumMatchVariantAndInactiveScopeImpl<T>)
where
    T::Value: EnumValue<Type = T>;

impl<T: EnumType> MatchVariantAndInactiveScope for EnumMatchVariantAndInactiveScope<T>
where
    T::Value: EnumValue<Type = T>,
{
    type MatchVariant = T::MatchVariant;
    type MatchActiveScope = Scope;

    fn match_activate_scope(self) -> (Self::MatchVariant, Self::MatchActiveScope) {
        T::match_activate_scope(self)
    }
}

impl<T: EnumType> EnumMatchVariantAndInactiveScope<T>
where
    T::Value: EnumValue<Type = T>,
{
    pub fn variant_access(&self) -> Interned<VariantAccess<T, Interned<dyn DynCanonicalType>>> {
        self.0.variant_access()
    }
    pub fn activate(
        self,
    ) -> (
        Interned<VariantAccess<T, Interned<dyn DynCanonicalType>>>,
        Scope,
    ) {
        self.0.activate()
    }
}

#[derive(Clone)]
pub struct EnumMatchVariantsIter<T: EnumType>
where
    T::Value: EnumValue<Type = T>,
{
    pub(crate) inner: EnumMatchVariantsIterImpl<T>,
    pub(crate) variant_index: std::ops::Range<usize>,
}

impl<T: EnumType> Iterator for EnumMatchVariantsIter<T>
where
    T::Value: EnumValue<Type = T>,
{
    type Item = EnumMatchVariantAndInactiveScope<T>;

    fn next(&mut self) -> Option<Self::Item> {
        self.variant_index.next().map(|variant_index| {
            EnumMatchVariantAndInactiveScope(self.inner.for_variant_index(variant_index))
        })
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        self.variant_index.size_hint()
    }
}

impl<T: EnumType> ExactSizeIterator for EnumMatchVariantsIter<T>
where
    T::Value: EnumValue<Type = T>,
{
    fn len(&self) -> usize {
        self.variant_index.len()
    }
}

impl<T: EnumType> FusedIterator for EnumMatchVariantsIter<T> where T::Value: EnumValue<Type = T> {}

impl<T: EnumType> DoubleEndedIterator for EnumMatchVariantsIter<T>
where
    T::Value: EnumValue<Type = T>,
{
    fn next_back(&mut self) -> Option<Self::Item> {
        self.variant_index.next_back().map(|variant_index| {
            EnumMatchVariantAndInactiveScope(self.inner.for_variant_index(variant_index))
        })
    }
}

impl Type for DynEnumType {
    type CanonicalType = DynEnumType;
    type Value = DynEnum;
    type CanonicalValue = DynEnum;
    type MaskType = UIntType<1>;
    type MaskValue = UInt<1>;
    type MatchVariant = Option<Expr<DynCanonicalValue>>;
    type MatchActiveScope = Scope;
    type MatchVariantAndInactiveScope = EnumMatchVariantAndInactiveScope<Self>;
    type MatchVariantsIter = EnumMatchVariantsIter<Self>;

    fn match_variants<IO: BundleValue>(
        this: Expr<Self::Value>,
        module_builder: &mut ModuleBuilder<IO, NormalModule>,
        source_location: SourceLocation,
    ) -> Self::MatchVariantsIter
    where
        IO::Type: crate::bundle::BundleType<Value = IO>,
    {
        module_builder.enum_match_variants_helper(this, source_location)
    }

    fn mask_type(&self) -> Self::MaskType {
        UIntType::new()
    }

    fn canonical(&self) -> Self::CanonicalType {
        *self
    }

    fn source_location(&self) -> SourceLocation {
        SourceLocation::builtin()
    }

    fn type_enum(&self) -> TypeEnum {
        TypeEnum::EnumType(*self)
    }

    fn from_canonical_type(t: Self::CanonicalType) -> Self {
        t
    }

    fn as_dyn_canonical_type_impl(this: &Self) -> Option<&dyn DynCanonicalType> {
        Some(this)
    }
}

impl Connect<Self> for DynEnumType {}

pub struct NoBuilder;

impl EnumType for DynEnumType {
    type Builder = NoBuilder;

    fn match_activate_scope(
        v: Self::MatchVariantAndInactiveScope,
    ) -> (Self::MatchVariant, Self::MatchActiveScope) {
        let (expr, scope) = v.0.activate();
        (expr.variant_type().ty.map(|_| expr.to_expr()), scope)
    }

    fn builder() -> Self::Builder {
        NoBuilder
    }

    fn variants(&self) -> Interned<[VariantType<Interned<dyn DynCanonicalType>>]> {
        self.0.variants
    }

    fn variants_hint() -> VariantsHint {
        VariantsHint {
            known_variants: [][..].intern(),
            more_variants: true,
        }
    }
}

impl CanonicalType for DynEnumType {
    const CANONICAL_TYPE_KIND: CanonicalTypeKind = CanonicalTypeKind::EnumType;
}

impl ToExpr for DynEnum {
    type Type = DynEnumType;

    fn ty(&self) -> Self::Type {
        self.ty
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        Expr::from_value(self)
    }
}

impl Value for DynEnum {
    fn to_canonical(&self) -> <Self::Type as Type>::CanonicalValue {
        self.clone()
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        this.ty
            .variant_to_bits(this.variant_index, this.variant_value.as_ref())
            .unwrap()
    }
}

impl EnumValue for DynEnum {}

impl CanonicalValue for DynEnum {
    fn value_enum_impl(this: &Self) -> ValueEnum {
        ValueEnum::Enum(this.clone())
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        this.ty
            .variant_to_bits(this.variant_index, this.variant_value.as_ref())
            .unwrap()
    }
}

mod impl_option {
    #[allow(dead_code)]
    #[derive(crate::Value)]
    #[hdl(target(std::option::Option), connect_inexact, outline_generated)]
    pub enum Option<T> {
        None,
        Some(T),
    }
}
