// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    array::ArrayType,
    bundle::{BundleType, BundleValue, DynBundle, DynBundleType, FieldType},
    enum_::{DynEnumType, EnumType, EnumValue},
    int::{DynSIntType, DynUIntType, FixedOrDynIntType, IntValue, UInt, UIntType},
    intern::{Intern, Interned, InternedCompare, PtrEqWithTypeId, SupportsPtrEqWithTypeId},
    memory::{DynPortType, MemPort, PortType},
    module::{
        transform::visit::{Fold, Folder, Visit, Visitor},
        Instance, ModuleIO, TargetName,
    },
    reg::Reg,
    source_location::SourceLocation,
    ty::{
        DynCanonicalType, DynCanonicalValue, DynType, DynValue, DynValueTrait, Type, TypeWithDeref,
        Value,
    },
    util::ConstBool,
    valueless::Valueless,
    wire::Wire,
};
use bitvec::slice::BitSlice;
use std::{any::Any, convert::Infallible, fmt, hash::Hash, marker::PhantomData, ops::Deref};

pub mod ops;

macro_rules! expr_enum {
    (
        pub enum $ExprEnum:ident {
            $($Variant:ident($VariantTy:ty),)+
        }
    ) => {
        #[derive(Copy, Clone, Eq, PartialEq, Hash)]
        pub enum $ExprEnum {
            $($Variant(Interned<$VariantTy>),)+
        }

        impl fmt::Debug for $ExprEnum {
            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
                match self {
                    $(Self::$Variant(v) => v.fmt(f),)+
                }
            }
        }

        impl $ExprEnum {
            pub fn target(self) -> Option<Interned<Target>> {
                match self {
                    $(Self::$Variant(v) => v.target(),)+
                }
            }
            #[allow(clippy::result_unit_err)]
            pub fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
                match self {
                    $(Self::$Variant(v) => v.to_literal_bits(),)+
                }
            }
        }

        impl ToExpr for $ExprEnum {
            type Type = Interned<dyn DynCanonicalType>;

            fn ty(&self) -> Self::Type {
                match self {
                    $(Self::$Variant(v) => v.ty().canonical_dyn(),)+
                }
            }

            fn to_expr(&self) -> Expr<DynCanonicalValue> {
                Expr::new_unchecked(*self)
            }
        }

        impl<State: ?Sized + Folder> Fold<State> for $ExprEnum {
            fn fold(self, state: &mut State) -> Result<Self, State::Error> {
                state.fold_expr_enum(self)
            }
            fn default_fold(self, state: &mut State) -> Result<Self, State::Error> {
                match self {
                    $(Self::$Variant(v) => Fold::fold(v, state).map(Self::$Variant),)+
                }
            }
        }

        impl<State: ?Sized + Visitor> Visit<State> for $ExprEnum {
            fn visit(&self, state: &mut State) -> Result<(), State::Error> {
                state.visit_expr_enum(self)
            }
            fn default_visit(&self, state: &mut State) -> Result<(), State::Error> {
                match self {
                    $(Self::$Variant(v) => Visit::visit(v, state),)+
                }
            }
        }
    };
}

expr_enum! {
    pub enum ExprEnum {
        Literal(Literal<Interned<dyn DynCanonicalType>>),
        ArrayLiteral(ops::ArrayLiteral<ArrayType<[DynCanonicalValue]>>),
        BundleLiteral(ops::BundleLiteral<DynBundleType>),
        EnumLiteral(ops::EnumLiteral<DynEnumType>),
        NotU(ops::Not<DynUIntType>),
        NotS(ops::Not<DynSIntType>),
        Neg(ops::Neg<DynSIntType>),
        BitAndU(ops::BitAnd<DynUIntType, DynUIntType>),
        BitAndS(ops::BitAnd<DynSIntType, DynSIntType>),
        BitOrU(ops::BitOr<DynUIntType, DynUIntType>),
        BitOrS(ops::BitOr<DynSIntType, DynSIntType>),
        BitXorU(ops::BitXor<DynUIntType, DynUIntType>),
        BitXorS(ops::BitXor<DynSIntType, DynSIntType>),
        AddU(ops::Add<DynUIntType, DynUIntType>),
        AddS(ops::Add<DynSIntType, DynSIntType>),
        SubU(ops::Sub<DynUIntType, DynUIntType>),
        SubS(ops::Sub<DynSIntType, DynSIntType>),
        MulU(ops::Mul<DynUIntType, DynUIntType>),
        MulS(ops::Mul<DynSIntType, DynSIntType>),
        DynShlU(ops::DynShl<DynUIntType>),
        DynShlS(ops::DynShl<DynSIntType>),
        DynShrU(ops::DynShr<DynUIntType>),
        DynShrS(ops::DynShr<DynSIntType>),
        FixedShlU(ops::FixedShl<DynUIntType>),
        FixedShlS(ops::FixedShl<DynSIntType>),
        FixedShrU(ops::FixedShr<DynUIntType>),
        FixedShrS(ops::FixedShr<DynSIntType>),
        CmpLtU(ops::CmpLt<DynUIntType, DynUIntType, DynUIntType>),
        CmpLtS(ops::CmpLt<DynSIntType, DynSIntType, DynUIntType>),
        CmpLeU(ops::CmpLe<DynUIntType, DynUIntType, DynUIntType>),
        CmpLeS(ops::CmpLe<DynSIntType, DynSIntType, DynUIntType>),
        CmpGtU(ops::CmpGt<DynUIntType, DynUIntType, DynUIntType>),
        CmpGtS(ops::CmpGt<DynSIntType, DynSIntType, DynUIntType>),
        CmpGeU(ops::CmpGe<DynUIntType, DynUIntType, DynUIntType>),
        CmpGeS(ops::CmpGe<DynSIntType, DynSIntType, DynUIntType>),
        CmpEqU(ops::CmpEq<DynUIntType, DynUIntType, DynUIntType>),
        CmpEqS(ops::CmpEq<DynSIntType, DynSIntType, DynUIntType>),
        CmpNeU(ops::CmpNe<DynUIntType, DynUIntType, DynUIntType>),
        CmpNeS(ops::CmpNe<DynSIntType, DynSIntType, DynUIntType>),
        CastUIntToUInt(ops::CastInt<DynUIntType, DynUIntType>),
        CastUIntToSInt(ops::CastInt<DynUIntType, DynSIntType>),
        CastSIntToUInt(ops::CastInt<DynSIntType, DynUIntType>),
        CastSIntToSInt(ops::CastInt<DynSIntType, DynSIntType>),
        SliceUInt(ops::Slice<DynUIntType>),
        SliceSInt(ops::Slice<DynSIntType>),
        ReduceBitAnd(ops::ReduceBitAnd<UIntType<1>>),
        ReduceBitOr(ops::ReduceBitOr<UIntType<1>>),
        ReduceBitXor(ops::ReduceBitXor<UIntType<1>>),
        FieldAccess(ops::FieldAccess<DynBundleType, Interned<dyn DynCanonicalType>>),
        VariantAccess(ops::VariantAccess<DynEnumType, Interned<dyn DynCanonicalType>>),
        ArrayIndex(ops::ArrayIndex<Interned<dyn DynType>>),
        DynArrayIndex(ops::DynArrayIndex<Interned<dyn DynType>>),
        CastToBits(ops::CastToBits),
        CastBitsTo(ops::CastBitsTo<Interned<dyn DynCanonicalType>>),
        CastBitToClock(ops::CastBitToClock),
        CastBitToSyncReset(ops::CastBitToSyncReset),
        CastBitToAsyncReset(ops::CastBitToAsyncReset),
        CastSyncResetToReset(ops::CastSyncResetToReset),
        CastAsyncResetToReset(ops::CastAsyncResetToReset),
        CastClockToBit(ops::CastClockToBit<DynUIntType>),
        CastSyncResetToBit(ops::CastSyncResetToBit<DynUIntType>),
        CastAsyncResetToBit(ops::CastAsyncResetToBit<DynUIntType>),
        CastResetToBit(ops::CastResetToBit<DynUIntType>),
        ModuleIO(ModuleIO<Interned<dyn DynCanonicalType>>),
        Instance(Instance<DynBundle>),
        Wire(Wire<Interned<dyn DynCanonicalType>>),
        Reg(Reg<Interned<dyn DynCanonicalType>>),
        MemPort(MemPort<DynPortType>),
    }
}

pub struct Expr<T> {
    /// use weird names to help work around rust-analyzer bug
    __enum: ExprEnum,
    __phantom: PhantomData<T>,
}

impl<T> Expr<T> {
    pub fn expr_enum(self) -> ExprEnum {
        self.__enum
    }
    pub fn new_unchecked(expr_enum: ExprEnum) -> Self {
        Self {
            __enum: expr_enum,
            __phantom: PhantomData,
        }
    }
    pub fn canonical(self) -> Expr<<T::Type as Type>::CanonicalValue>
    where
        T: Value,
    {
        Expr {
            __enum: self.expr_enum(),
            __phantom: PhantomData,
        }
    }
    pub fn dyn_canonical_type(self) -> Interned<dyn DynCanonicalType> {
        self.expr_enum().ty()
    }
    pub fn canonical_type(self) -> <T::Type as Type>::CanonicalType
    where
        T: Value,
    {
        <T::Type as Type>::CanonicalType::from_dyn_canonical_type(self.dyn_canonical_type())
    }
    pub fn valueless(self) -> Valueless<T::Type>
    where
        T: Value<Type: Type<Value = T>>,
    {
        Valueless { ty: self.ty() }
    }
    #[allow(clippy::result_unit_err)]
    pub fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
        self.expr_enum().to_literal_bits()
    }
    #[track_caller]
    pub fn with_type<ToValue: Value<Type: Type<Value = ToValue>>>(self) -> Expr<ToValue>
    where
        T: Value<Type: Type<Value = T>>,
    {
        let retval = Expr::<ToValue>::new_unchecked(self.expr_enum());
        let _ = retval.ty(); // check that the type is correct
        retval
    }
    pub fn to_dyn(self) -> Expr<DynValue> {
        Expr::new_unchecked(self.expr_enum())
    }
    pub fn to_canonical_dyn(self) -> Expr<DynCanonicalValue> {
        Expr::new_unchecked(self.expr_enum())
    }
    #[track_caller]
    pub fn from_value(value: &T) -> Self
    where
        T: Value<Type: Type<Value = T>>,
    {
        Literal::<T::Type>::new_unchecked(value.to_canonical()).to_expr()
    }
    pub fn target(self) -> Option<Interned<Target>> {
        self.expr_enum().target()
    }
    pub fn flow(self) -> Flow {
        self.target().map(|v| v.flow()).unwrap_or(Flow::Source)
    }
}

impl<T> Copy for Expr<T> {}

impl<T> Clone for Expr<T> {
    fn clone(&self) -> Self {
        *self
    }
}

impl<T> fmt::Debug for Expr<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            __enum,
            __phantom: _,
        } = self;
        __enum.fmt(f)
    }
}

impl<T: Value<Type: Type<Value = T>>> sealed::Sealed for Expr<T> {}

impl<T> PartialEq for Expr<T> {
    fn eq(&self, other: &Self) -> bool {
        let Self {
            __enum,
            __phantom: _,
        } = self;
        *__enum == other.__enum
    }
}

impl<T> Eq for Expr<T> {}

impl<T> Hash for Expr<T> {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        let Self {
            __enum,
            __phantom: _,
        } = self;
        __enum.hash(state);
    }
}

impl<T> Expr<IntValue<T>>
where
    T: FixedOrDynIntType<1, Signed = ConstBool<false>>,
{
    pub fn as_bool(self) -> Expr<UInt<1>> {
        assert_eq!(self.canonical_type().width, 1);
        Expr::new_unchecked(self.expr_enum())
    }
}

impl<T: BundleValue<Type: BundleType<Value = T>>> Expr<T> {
    pub fn field<FieldValue: Value<Type: Type<Value = FieldValue>>>(
        self,
        name: &str,
    ) -> Expr<FieldValue> {
        ops::FieldAccess::<T::Type, FieldValue::Type>::new_unchecked(
            self.canonical(),
            name.intern(),
        )
        .to_expr()
    }
}

impl<T: Value<Type: Type<Value = T>>> ToExpr for Expr<T> {
    type Type = T::Type;

    fn ty(&self) -> T::Type {
        T::Type::from_dyn_canonical_type(self.dyn_canonical_type())
    }

    fn to_expr(&self) -> Expr<T> {
        *self
    }
}

impl<V: Value<Type = T>, T> Deref for Expr<V>
where
    T: TypeWithDeref<Value = V>,
{
    type Target = T::MatchVariant;

    fn deref(&self) -> &Self::Target {
        T::expr_deref(self)
    }
}

impl<V: Value<Type: Type<Value = V>>, State: ?Sized + Folder> Fold<State> for Expr<V> {
    fn fold(self, state: &mut State) -> Result<Self, State::Error> {
        state.fold_expr(self)
    }
    fn default_fold(self, state: &mut State) -> Result<Self, State::Error> {
        Ok(Expr::<DynCanonicalValue>::new_unchecked(self.expr_enum().fold(state)?).with_type())
    }
}

impl<V: Value<Type: Type<Value = V>>, State: ?Sized + Visitor> Visit<State> for Expr<V> {
    fn visit(&self, state: &mut State) -> Result<(), State::Error> {
        state.visit_expr(self)
    }
    fn default_visit(&self, state: &mut State) -> Result<(), State::Error> {
        self.expr_enum().visit(state)
    }
}

pub trait ExprTraitBase:
    fmt::Debug + Any + SupportsPtrEqWithTypeId + Send + Sync + sealed::Sealed + ToExpr
{
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash, Debug)]
pub enum Flow {
    Source,
    Sink,
    Duplex,
}

impl Flow {
    pub const fn flip(self) -> Flow {
        match self {
            Flow::Source => Flow::Sink,
            Flow::Sink => Flow::Source,
            Flow::Duplex => Flow::Duplex,
        }
    }
    pub const fn flip_if(self, flipped: bool) -> Flow {
        if flipped {
            self.flip()
        } else {
            self
        }
    }
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct TargetPathBundleField {
    pub name: Interned<str>,
}

impl fmt::Display for TargetPathBundleField {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, ".{}", self.name)
    }
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct TargetPathArrayElement {
    pub index: usize,
}

impl fmt::Display for TargetPathArrayElement {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "[{}]", self.index)
    }
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct TargetPathDynArrayElement {}

impl fmt::Display for TargetPathDynArrayElement {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "[<dyn>]")
    }
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum TargetPathElement {
    BundleField(TargetPathBundleField),
    ArrayElement(TargetPathArrayElement),
    DynArrayElement(TargetPathDynArrayElement),
}

impl From<TargetPathBundleField> for TargetPathElement {
    fn from(value: TargetPathBundleField) -> Self {
        Self::BundleField(value)
    }
}

impl From<TargetPathArrayElement> for TargetPathElement {
    fn from(value: TargetPathArrayElement) -> Self {
        Self::ArrayElement(value)
    }
}

impl From<TargetPathDynArrayElement> for TargetPathElement {
    fn from(value: TargetPathDynArrayElement) -> Self {
        Self::DynArrayElement(value)
    }
}

impl fmt::Display for TargetPathElement {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::BundleField(v) => v.fmt(f),
            Self::ArrayElement(v) => v.fmt(f),
            Self::DynArrayElement(v) => v.fmt(f),
        }
    }
}

impl TargetPathElement {
    pub fn canonical_ty(&self, parent: Interned<Target>) -> Interned<dyn DynCanonicalType> {
        match self {
            &Self::BundleField(TargetPathBundleField { name }) => {
                let parent_ty = parent
                    .canonical_ty()
                    .type_enum()
                    .bundle_type()
                    .expect("parent type is known to be a bundle");
                let field = parent_ty
                    .field_by_name(name)
                    .expect("field name is known to be a valid field of parent type");
                field.ty
            }
            &Self::ArrayElement(TargetPathArrayElement { index }) => {
                let parent_ty = parent
                    .canonical_ty()
                    .type_enum()
                    .array_type()
                    .expect("parent type is known to be an array");
                assert!(index < parent_ty.len());
                *parent_ty.element()
            }
            Self::DynArrayElement(_) => {
                let parent_ty = parent
                    .canonical_ty()
                    .type_enum()
                    .array_type()
                    .expect("parent type is known to be an array");
                *parent_ty.element()
            }
        }
    }
    pub fn flow(&self, parent: Interned<Target>) -> Flow {
        match self {
            Self::BundleField(v) => {
                let parent_ty = parent
                    .canonical_ty()
                    .type_enum()
                    .bundle_type()
                    .expect("parent type is known to be a bundle");
                let field = parent_ty
                    .field_by_name(v.name)
                    .expect("field name is known to be a valid field of parent type");
                parent.flow().flip_if(field.flipped)
            }
            Self::ArrayElement(_) => parent.flow(),
            Self::DynArrayElement(_) => parent.flow(),
        }
    }
    pub fn is_static(&self) -> bool {
        match self {
            Self::BundleField(_) | Self::ArrayElement(_) => true,
            Self::DynArrayElement(_) => false,
        }
    }
}

macro_rules! impl_target_base {
    (
        $(#[$enum_meta:meta])*
        $enum_vis:vis enum $TargetBase:ident {
            $(
                #[is = $is_fn:ident]
                #[to = $to_fn:ident]
                $(#[$variant_meta:meta])*
                $Variant:ident($VariantTy:ty),
            )*
        }
    ) => {
        $(#[$enum_meta])*
        $enum_vis enum $TargetBase {
            $(
                $(#[$variant_meta])*
                $Variant($VariantTy),
            )*
        }

        impl fmt::Debug for $TargetBase {
            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
                match self {
                    $(Self::$Variant(v) => v.fmt(f),)*
                }
            }
        }

        $(
        impl From<$VariantTy> for $TargetBase {
            fn from(value: $VariantTy) -> Self {
                Self::$Variant(value)
            }
        }

        impl From<$VariantTy> for Target {
            fn from(value: $VariantTy) -> Self {
                $TargetBase::$Variant(value).into()
            }
        }
        )*

        impl $TargetBase {
            $(
            $enum_vis fn $is_fn(&self) -> bool {
                self.$to_fn().is_some()
            }
            $enum_vis fn $to_fn(&self) -> Option<&$VariantTy> {
                if let Self::$Variant(retval) = self {
                    Some(retval)
                } else {
                    None
                }
            }
            )*
            $enum_vis fn must_connect_to(&self) -> bool {
                match self {
                    $(Self::$Variant(v) => v.must_connect_to(),)*
                }
            }
            $enum_vis fn flow(&self) -> Flow {
                match self {
                    $(Self::$Variant(v) => v.flow(),)*
                }
            }
            $enum_vis fn source_location(&self) -> SourceLocation {
                match self {
                    $(Self::$Variant(v) => v.source_location(),)*
                }
            }
        }
    };
}

impl_target_base! {
    #[derive(Clone, PartialEq, Eq, Hash)]
    pub enum TargetBase {
        #[is = is_module_io]
        #[to = module_io]
        ModuleIO(ModuleIO<Interned<dyn DynCanonicalType>>),
        #[is = is_mem_port]
        #[to = mem_port]
        MemPort(MemPort<DynPortType>),
        #[is = is_reg]
        #[to = reg]
        Reg(Reg<Interned<dyn DynCanonicalType>>),
        #[is = is_wire]
        #[to = wire]
        Wire(Wire<Interned<dyn DynCanonicalType>>),
        #[is = is_instance]
        #[to = instance]
        Instance(Instance<DynBundle>),
    }
}

impl fmt::Display for TargetBase {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{:?}", self.target_name())
    }
}

impl TargetBase {
    pub fn target_name(&self) -> TargetName {
        match self {
            TargetBase::ModuleIO(v) => TargetName(v.scoped_name(), None),
            TargetBase::MemPort(v) => TargetName(v.mem_name(), Some(v.port_name())),
            TargetBase::Reg(v) => TargetName(v.scoped_name(), None),
            TargetBase::Wire(v) => TargetName(v.scoped_name(), None),
            TargetBase::Instance(v) => TargetName(v.scoped_name(), None),
        }
    }
    pub fn canonical_ty(&self) -> Interned<dyn DynCanonicalType> {
        match self {
            TargetBase::ModuleIO(v) => v.ty(),
            TargetBase::MemPort(v) => v.ty().canonical_dyn(),
            TargetBase::Reg(v) => v.ty(),
            TargetBase::Wire(v) => v.ty(),
            TargetBase::Instance(v) => v.ty().canonical_dyn(),
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub struct TargetChild {
    parent: Interned<Target>,
    path_element: Interned<TargetPathElement>,
    canonical_ty: Interned<dyn DynCanonicalType>,
    flow: Flow,
}

impl fmt::Debug for TargetChild {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            parent,
            path_element,
            canonical_ty: _,
            flow: _,
        } = self;
        parent.fmt(f)?;
        fmt::Display::fmt(path_element, f)
    }
}

impl fmt::Display for TargetChild {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            parent,
            path_element,
            canonical_ty: _,
            flow: _,
        } = self;
        parent.fmt(f)?;
        path_element.fmt(f)
    }
}

impl TargetChild {
    pub fn new(parent: Interned<Target>, path_element: Interned<TargetPathElement>) -> Self {
        Self {
            parent,
            path_element,
            canonical_ty: path_element.canonical_ty(parent),
            flow: path_element.flow(parent),
        }
    }
    pub fn parent(self) -> Interned<Target> {
        self.parent
    }
    pub fn path_element(self) -> Interned<TargetPathElement> {
        self.path_element
    }
    pub fn canonical_ty(self) -> Interned<dyn DynCanonicalType> {
        self.canonical_ty
    }
    pub fn flow(self) -> Flow {
        self.flow
    }
    pub fn bundle_field(self) -> Option<FieldType<Interned<dyn DynCanonicalType>>> {
        if let TargetPathElement::BundleField(TargetPathBundleField { name }) = *self.path_element {
            let parent_ty = self
                .parent
                .canonical_ty()
                .type_enum()
                .bundle_type()
                .expect("parent known to be bundle");
            Some(
                parent_ty
                    .field_by_name(name)
                    .expect("field name known to be a valid field of parent"),
            )
        } else {
            None
        }
    }
}

#[derive(Clone, PartialEq, Eq, Hash)]
pub enum Target {
    Base(Interned<TargetBase>),
    Child(TargetChild),
}

impl From<TargetBase> for Target {
    fn from(value: TargetBase) -> Self {
        Self::Base(Intern::intern_sized(value))
    }
}

impl From<TargetChild> for Target {
    fn from(value: TargetChild) -> Self {
        Self::Child(value)
    }
}

impl From<Interned<TargetBase>> for Target {
    fn from(value: Interned<TargetBase>) -> Self {
        Self::Base(value)
    }
}

impl Target {
    pub fn base(&self) -> Interned<TargetBase> {
        let mut target = self;
        loop {
            match target {
                Self::Base(v) => break *v,
                Self::Child(v) => target = &v.parent,
            }
        }
    }
    pub fn child(&self) -> Option<TargetChild> {
        match *self {
            Target::Base(_) => None,
            Target::Child(v) => Some(v),
        }
    }
    pub fn is_static(&self) -> bool {
        let mut target = self;
        loop {
            match target {
                Self::Base(_) => return true,
                Self::Child(v) if !v.path_element().is_static() => return false,
                Self::Child(v) => target = &v.parent,
            }
        }
    }
    #[must_use]
    pub fn join(&self, path_element: Interned<TargetPathElement>) -> Self {
        TargetChild::new(self.intern(), path_element).into()
    }
    pub fn flow(&self) -> Flow {
        match self {
            Self::Base(v) => v.flow(),
            Self::Child(v) => v.flow(),
        }
    }
    pub fn canonical_ty(&self) -> Interned<dyn DynCanonicalType> {
        match self {
            Target::Base(v) => v.canonical_ty(),
            Target::Child(v) => v.canonical_ty(),
        }
    }
}

impl fmt::Display for Target {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Base(v) => v.fmt(f),
            Self::Child(v) => v.fmt(f),
        }
    }
}

impl fmt::Debug for Target {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Base(v) => v.fmt(f),
            Self::Child(v) => v.fmt(f),
        }
    }
}

pub trait ExprTrait: ExprTraitBase {
    fn expr_enum(&self) -> ExprEnum;
    fn target(&self) -> Option<Interned<Target>>;
    fn valueless(&self) -> Valueless<Self::Type> {
        Valueless { ty: self.ty() }
    }
    #[allow(clippy::result_unit_err)]
    fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()>;
}

impl<T: ExprTrait> ExprTraitBase for T {}

impl<T: Type> InternedCompare for dyn ExprTrait<Type = T> {
    type InternedCompareKey = PtrEqWithTypeId;
    fn interned_compare_key_ref(this: &Self) -> Self::InternedCompareKey {
        Self::get_ptr_eq_with_type_id(this)
    }
    fn interned_compare_key_weak(this: &std::sync::Weak<Self>) -> Self::InternedCompareKey {
        Self::get_ptr_eq_with_type_id(&*this.upgrade().unwrap())
    }
}

pub struct SimState {}

mod sealed {
    pub trait Sealed {}
}

pub trait ToExpr {
    type Type: Type;
    fn ty(&self) -> Self::Type;
    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value>;
}

impl<T: ?Sized + ToExpr> ToExpr for &'_ T {
    type Type = T::Type;

    fn ty(&self) -> Self::Type {
        (**self).ty()
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        (**self).to_expr()
    }
}

impl<T: ?Sized + ToExpr> ToExpr for &'_ mut T {
    type Type = T::Type;

    fn ty(&self) -> Self::Type {
        (**self).ty()
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        (**self).to_expr()
    }
}

impl<T: ?Sized + ToExpr> ToExpr for Box<T> {
    type Type = T::Type;

    fn ty(&self) -> Self::Type {
        (**self).ty()
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        (**self).to_expr()
    }
}

#[derive(Clone, Eq, PartialEq, Hash)]
pub struct Literal<T: Type> {
    value: T::CanonicalValue,
}

impl<T: Type + fmt::Debug> fmt::Debug for Literal<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.value.fmt(f)
    }
}

impl<T: Type> Literal<T> {
    #[track_caller]
    pub fn new_unchecked(value: T::CanonicalValue) -> Self {
        assert!(
            value.ty().is_passive(),
            "can't have a literal with flipped fields"
        );
        Self { value }
    }
    pub fn value(&self) -> &T::CanonicalValue {
        &self.value
    }
    pub fn canonical(&self) -> Literal<T::CanonicalType> {
        Literal {
            value: self.value.clone(),
        }
    }
}

impl<T: Type> sealed::Sealed for Literal<T> {}

impl<T: Type> ToExpr for Literal<T> {
    type Type = T;

    fn ty(&self) -> Self::Type {
        Self::Type::from_canonical_type(self.value.ty())
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        Expr::new_unchecked(self.expr_enum())
    }
}

impl<T: Type> ExprTrait for Literal<T> {
    fn expr_enum(&self) -> ExprEnum {
        ExprEnum::Literal(
            Literal {
                value: self.value.to_canonical_dyn(),
            }
            .intern_sized(),
        )
    }

    fn target(&self) -> Option<Interned<Target>> {
        None
    }

    fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
        Ok(self.value.to_bits())
    }
}

impl<T: Type, State: ?Sized + Folder> Fold<State> for Literal<T>
where
    T::CanonicalValue: Fold<State>,
{
    fn fold(self, state: &mut State) -> Result<Self, State::Error> {
        state.fold_literal(self)
    }
    fn default_fold(self, state: &mut State) -> Result<Self, State::Error> {
        Ok(Literal {
            value: self.value.fold(state)?,
        })
    }
}

impl<T: Type, State: ?Sized + Visitor> Visit<State> for Literal<T>
where
    T::CanonicalValue: Visit<State>,
{
    fn visit(&self, state: &mut State) -> Result<(), <State as Visitor>::Error> {
        state.visit_literal(self)
    }
    fn default_visit(&self, state: &mut State) -> Result<(), <State as Visitor>::Error> {
        self.value.visit(state)
    }
}

impl<T: Type> sealed::Sealed for ModuleIO<T> {}

impl<T: Type> ToExpr for ModuleIO<T> {
    type Type = T;

    fn ty(&self) -> Self::Type {
        self.field_type().ty.clone()
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        Expr::new_unchecked(self.expr_enum())
    }
}

impl<T: Type> ExprTrait for ModuleIO<T> {
    fn expr_enum(&self) -> ExprEnum {
        ExprEnum::ModuleIO(self.to_canonical_dyn_module_io().intern_sized())
    }

    fn target(&self) -> Option<Interned<Target>> {
        Some(Intern::intern_sized(
            self.to_canonical_dyn_module_io().into(),
        ))
    }

    fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
        Err(())
    }
}

impl<T: BundleValue> sealed::Sealed for Instance<T> where T::Type: BundleType<Value = T> {}

impl<T: BundleValue> ToExpr for Instance<T>
where
    T::Type: BundleType<Value = T>,
{
    type Type = T::Type;

    fn ty(&self) -> Self::Type {
        (*self.instantiated().io_ty()).clone()
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        Expr::new_unchecked(self.expr_enum())
    }
}

impl<T: BundleValue> ExprTrait for Instance<T>
where
    T::Type: BundleType<Value = T>,
{
    fn expr_enum(&self) -> ExprEnum {
        ExprEnum::Instance(self.canonical().intern_sized())
    }

    fn target(&self) -> Option<Interned<Target>> {
        Some(Intern::intern_sized(self.canonical().into()))
    }

    fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
        Err(())
    }
}

impl<T: Type> sealed::Sealed for Wire<T> {}

impl<T: Type> ExprTrait for Wire<T> {
    fn expr_enum(&self) -> ExprEnum {
        ExprEnum::Wire(self.to_dyn_canonical_wire().intern_sized())
    }

    fn target(&self) -> Option<Interned<Target>> {
        Some(Intern::intern_sized(self.to_dyn_canonical_wire().into()))
    }

    fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
        Err(())
    }
}

impl<T: Type> sealed::Sealed for Reg<T> {}

impl<T: Type> ExprTrait for Reg<T> {
    fn expr_enum(&self) -> ExprEnum {
        ExprEnum::Reg(self.to_dyn_canonical_reg().intern_sized())
    }

    fn target(&self) -> Option<Interned<Target>> {
        Some(Intern::intern_sized(self.to_dyn_canonical_reg().into()))
    }

    fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
        Err(())
    }
}

#[doc(hidden)]
pub fn value_from_expr_type<V: Value>(_expr: Expr<V>, infallible: Infallible) -> V {
    match infallible {}
}

#[doc(hidden)]
pub fn check_match_expr<V: Value>(_expr: Expr<V>, _check_fn: impl FnOnce(V, Infallible)) {}

#[doc(hidden)]
#[inline]
pub fn make_enum_expr<V: EnumValue>(
    _check_fn: impl FnOnce(Infallible) -> V,
    build: impl FnOnce(<V::Type as EnumType>::Builder) -> Expr<V>,
) -> Expr<V>
where
    V::Type: EnumType<Value = V>,
{
    build(V::Type::builder())
}

#[doc(hidden)]
#[inline]
pub fn make_bundle_expr<V: BundleValue>(
    _check_fn: impl FnOnce(Infallible) -> V,
    build: impl FnOnce(<V::Type as BundleType>::Builder) -> Expr<V>,
) -> Expr<V>
where
    V::Type: BundleType<Value = V>,
{
    build(V::Type::builder())
}

impl<T: PortType> sealed::Sealed for MemPort<T> where Self: ToExpr<Type = T::PortType> {}

impl<T: PortType> ExprTrait for MemPort<T>
where
    Self: ToExpr<Type = T::PortType>,
{
    fn expr_enum(&self) -> ExprEnum {
        ExprEnum::MemPort(self.canonical().intern_sized())
    }
    fn target(&self) -> Option<Interned<Target>> {
        Some(Intern::intern_sized(self.canonical().into()))
    }
    fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
        Err(())
    }
}
