// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    array::{Array, ArrayType, ArrayTypeTrait, ValueArrayOrSlice},
    bundle::{BundleType, BundleValue, DynBundleType, FieldType},
    clock::{Clock, ClockType, ToClock},
    enum_::{DynEnumType, EnumType, EnumValue, VariantType},
    expr::{
        sealed, Expr, ExprEnum, ExprTrait, Target, TargetPathArrayElement, TargetPathBundleField,
        TargetPathDynArrayElement, TargetPathElement, ToExpr,
    },
    int::{
        DynIntType, DynSInt, DynSIntType, DynUInt, DynUIntType, FixedOrDynIntType, Int, IntCmp,
        IntTypeTrait, IntValue, UInt, UIntType,
    },
    intern::{Intern, Interned},
    reset::{
        AsyncReset, AsyncResetType, Reset, ResetType, SyncReset, SyncResetType, ToAsyncReset,
        ToReset, ToSyncReset,
    },
    ty::{
        CanonicalType, CanonicalValue, DynCanonicalType, DynCanonicalValue, DynType, DynValueTrait,
        Type, Value,
    },
    util::{interned_bit, ConstBool, ConstBoolDispatch, ConstBoolDispatchTag, GenericConstBool},
    valueless::{Valueless, ValuelessTr},
};
use bitvec::{slice::BitSlice, vec::BitVec};
use num_traits::ToPrimitive;
use std::{
    fmt,
    hash::{Hash, Hasher},
    ops::{self, Index, Range, RangeBounds},
};

macro_rules! fixed_ary_op {
    (
        pub struct $name:ident<$($T:ident,)* $(#[const] $C:ident: $CTy:ty,)*>
        where
            ($($where:tt)*)
        {
            $($arg_vis:vis $arg:ident: $Arg:ty,)+
            $(#[cache]
            $cache_before_ty:ident: $CacheBeforeTy:ty = $cache_before_ty_expr:expr,)*
            #[type$(($ty_arg:ident))?]
            $ty_vis:vis $ty_name:ident: $Ty:ty = $ty_expr:expr,
            $(#[cache]
            $cache_after_ty:ident: $CacheAfterTy:ty = $cache_after_ty_expr:expr,)*
            $(#[target]
            $target_name:ident: Option<Interned<Target>> = $target_expr:expr,)?
            fn simulate(&$simulate_self:ident, $sim_state:ident: &mut SimState) -> _ {
                $($simulate_body:tt)+
            }

            fn expr_enum(&$expr_enum_self:ident) -> _ {
                $($expr_enum_body:tt)+
            }

            fn to_literal_bits(&$to_literal_bits_self:ident) -> Result<Interned<BitSlice>, ()> {
                $($to_literal_bits_body:tt)+
            }
        }
    ) => {
        pub struct $name<$($T,)* $(const $C: $CTy,)*>
        where
            $($where)*
        {
            $($arg_vis $arg: $Arg,)+
            $($cache_before_ty: $CacheBeforeTy,)*
            $ty_vis $ty_name: $Ty,
            $($cache_after_ty: $CacheAfterTy,)*
            $($target_name: Option<Interned<Target>>,)?
        }

        impl<$($T,)* $(const $C: $CTy,)*> $name<$($T,)* $($C,)*>
        where
            $($where)*
        {
            pub fn new_unchecked($($arg: $Arg,)* $($ty_arg: $Ty,)?) -> Self {
                $(let $cache_before_ty: $CacheBeforeTy = $cache_before_ty_expr;)*
                let $ty_name: $Ty = $ty_expr;
                $(let $cache_after_ty: $CacheAfterTy = $cache_after_ty_expr;)*
                $(let $target_name: Option<Interned<Target>> = $target_expr;)?
                Self {
                    $($arg,)*
                    $($cache_before_ty,)*
                    $ty_name,
                    $($cache_after_ty,)*
                    $($target_name,)?
                }
            }
            pub fn canonical(&self) -> $name<$($T::CanonicalType,)* $($C,)*> {
                $name {
                    $($arg: self.$arg.clone(),)*
                    $($cache_before_ty: self.$cache_before_ty.clone(),)*
                    $ty_name: self.$ty_name.canonical(),
                    $($cache_after_ty: self.$cache_after_ty.clone(),)*
                    $($target_name: self.$target_name,)?
                }
            }
        }

        impl<$($T,)* $(const $C: $CTy,)*> Clone for $name<$($T,)* $($C,)*>
        where
            $($where)*
        {
            fn clone(&self) -> Self {
                Self {
                    $($arg: self.$arg.clone(),)*
                    $($cache_before_ty: self.$cache_before_ty.clone(),)*
                    $ty_name: self.$ty_name.clone(),
                    $($cache_after_ty: self.$cache_after_ty.clone(),)*
                    $($target_name: self.$target_name,)?
                }
            }
        }

        impl<$($T,)* $(const $C: $CTy,)*> PartialEq for $name<$($T,)* $($C,)*>
        where
            $($where)*
        {
            fn eq(&self, rhs: &Self) -> bool {
                $(self.$arg == rhs.$arg &&)* self.$ty_name == rhs.$ty_name
            }
        }

        impl<$($T,)* $(const $C: $CTy,)*> Eq for $name<$($T,)* $($C,)*>
        where
            $($where)*
        {
        }

        impl<$($T,)* $(const $C: $CTy,)*> Hash for $name<$($T,)* $($C,)*>
        where
            $($where)*
        {
            fn hash<H: Hasher>(&self, state: &mut H) {
                $(self.$arg.hash(state);)*
                self.$ty_name.hash(state);
            }
        }

        impl<$($T,)* $(const $C: $CTy,)*> fmt::Debug for $name<$($T,)* $($C,)*>
        where
            $($where)*
        {
            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
                f.debug_tuple(stringify!($name))$(.field(&self.$arg))*.finish()
            }
        }

        impl<$($T,)* $(const $C: $CTy,)*> ToExpr for $name<$($T,)* $($C,)*>
        where
            $($where)*
        {
            type Type = $Ty;

            fn ty(&self) -> Self::Type {
                self.$ty_name.clone()
            }
            fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
                Expr::new_unchecked(self.expr_enum())
            }
        }

        impl<$($T,)* $(const $C: $CTy,)*> ExprTrait for $name<$($T,)* $($C,)*>
        where
            $($where)*
        {
            fn expr_enum(&$expr_enum_self) -> ExprEnum {
                $($expr_enum_body)+
            }
            fn target(&self) -> Option<Interned<Target>> {
                ($(self.$target_name,)? None::<Interned<Target>>,).0
            }
            fn to_literal_bits(&$to_literal_bits_self) -> Result<Interned<BitSlice>, ()> {
                $($to_literal_bits_body)+
            }
        }

        impl<$($T,)* $(const $C: $CTy,)*> sealed::Sealed for $name<$($T,)* $($C,)*>
        where
            $($where)*
        {
        }
    };
}

macro_rules! unary_op {
    (
        #[method = $op:ident]
        impl<$T:ident> $Op:ident for _ where ($($where:tt)*) {
            fn expr_enum(&$expr_enum_self:ident) -> _ {
                $($expr_enum_body:tt)+
            }
        }
    ) => {
        fixed_ary_op! {
            pub struct $Op<$T,>
            where (
                $($where)*
            )
            {
                pub arg: Expr<$T::CanonicalValue>,
                #[type]
                ty: <<Valueless<$T> as ops::$Op>::Output as ValuelessTr>::Type = ops::$Op::$op(
                    Valueless::<$T>::from_canonical(arg.valueless()),
                ).ty,
                #[cache]
                literal_bits: Result<Interned<BitSlice>, ()> = {
                    arg.to_literal_bits().map(|v| ops::$Op::$op($T::CanonicalValue::from_bit_slice(&v)).to_bits())
                },

                fn simulate(&self, sim_state: &mut SimState) -> _ {
                    ops::$Op::$op(self.arg.simulate(sim_state))
                }

                fn expr_enum(&$expr_enum_self) -> _ {
                    $($expr_enum_body)+
                }

                fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
                    self.literal_bits
                }
            }
        }

        impl<$T, V: Value<Type = T>> ops::$Op for Expr<V>
        where
            $($where)*
        {
            type Output = Expr<<<Valueless<$T> as ops::$Op>::Output as ValuelessTr>::Value>;

            fn $op(self) -> Self::Output {
                $Op::<$T>::new_unchecked(self.canonical()).to_expr()
            }
        }
    };
}

unary_op! {
    #[method = not]
    impl<T> Not for _
    where (
        T: IntTypeTrait,
    )
    {
        fn expr_enum(&self) -> _ {
            struct Tag;
            impl ConstBoolDispatchTag for Tag {
                type Type<Select: GenericConstBool> = Not<DynIntType<Select>>;
            }
            match ConstBoolDispatch::new::<Tag, T::Signed>(self.canonical()) {
                ConstBoolDispatch::False(v) => ExprEnum::NotU(v.intern_sized()),
                ConstBoolDispatch::True(v) => ExprEnum::NotS(v.intern_sized()),
            }
        }
    }
}

unary_op! {
    #[method = neg]
    impl<T> Neg for _
    where (
        T: IntTypeTrait<Signed = ConstBool<true>>,
    )
    {
        fn expr_enum(&self) -> _ {
            ExprEnum::Neg(self.canonical().intern_sized())
        }
    }
}

macro_rules! binary_op {
    (
        #[method = $op:ident, rhs_to_canonical_dyn = $rhs_to_canonical_dyn:ident, expr_enum_u = $expr_enum_u:ident, expr_enum_s = $expr_enum_s:ident]
        impl<$LhsType:ident, $RhsType:ident> $Op:ident for _ where $($where:tt)*
    ) => {
        fixed_ary_op! {
            pub struct $Op<$LhsType, $RhsType,>
            where (
                $($where)*
            )
            {
                pub lhs: Expr<$LhsType::CanonicalValue>,
                pub rhs: Expr<$RhsType::CanonicalValue>,
                #[type]
                ty: <<Valueless<$LhsType> as ops::$Op<Valueless<$RhsType>>>::Output as ValuelessTr>::Type = ops::$Op::$op(
                    Valueless::<$LhsType>::from_canonical(lhs.valueless()),
                    Valueless::<$RhsType>::from_canonical(rhs.valueless()),
                ).ty,
                #[cache]
                literal_bits: Result<Interned<BitSlice>, ()> = {
                    lhs.to_literal_bits()
                        .ok()
                        .zip(rhs.to_literal_bits().ok())
                        .map(|(lhs, rhs)| {
                            ops::$Op::$op(
                                $LhsType::CanonicalValue::from_bit_slice(&lhs),
                                $RhsType::CanonicalValue::from_bit_slice(&rhs),
                            )
                            .to_bits()
                        })
                        .ok_or(())
                },
                fn simulate(&self, sim_state: &mut SimState) -> _ {
                    ops::$Op::$op(self.lhs.simulate(sim_state), self.rhs.simulate(sim_state))
                }
                fn expr_enum(&self) -> _ {
                    struct Tag;
                    impl ConstBoolDispatchTag for Tag {
                        type Type<Select: GenericConstBool> = $Op<DynIntType<Select>, DynIntType<Select>>;
                    }
                    match ConstBoolDispatch::new::<Tag, $LhsType::Signed>(self.canonical()) {
                        ConstBoolDispatch::False(v) => ExprEnum::$expr_enum_u(v.intern_sized()),
                        ConstBoolDispatch::True(v) => ExprEnum::$expr_enum_s(v.intern_sized()),
                    }
                }
                fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
                    self.literal_bits
                }
            }
        }

        impl<Lhs: Value<Type = $LhsType>, $LhsType, $RhsType, Rhs: Value<Type = $RhsType>> ops::$Op<Expr<Rhs>> for Expr<Lhs>
        where
            $($where)*
        {
            type Output = Expr<<<Valueless<$LhsType> as ops::$Op<Valueless<$RhsType>>>::Output as ValuelessTr>::Value>;

            fn $op(self, rhs: Expr<Rhs>) -> Self::Output {
                $Op::<$LhsType, $RhsType>::new_unchecked(self.canonical(), rhs.canonical()).to_expr()
            }
        }

        impl<Lhs: Value<Type = $LhsType>, $LhsType, $RhsType, Rhs: Value<Type = $RhsType>> ops::$Op<Rhs> for Expr<Lhs>
        where
            $($where)*
        {
            type Output = Expr<<<Valueless<$LhsType> as ops::$Op<Valueless<$RhsType>>>::Output as ValuelessTr>::Value>;

            fn $op(self, rhs: Rhs) -> Self::Output {
                $Op::<$LhsType, $RhsType>::new_unchecked(self.canonical(), rhs.to_expr().canonical()).to_expr()
            }
        }

        impl<$LhsType, $RhsType, Rhs: Value<Type = $RhsType>> ops::$Op<Expr<Rhs>> for IntValue<$LhsType>
        where
            $($where)*
        {
            type Output = Expr<<<Valueless<$LhsType> as ops::$Op<Valueless<$RhsType>>>::Output as ValuelessTr>::Value>;

            fn $op(self, rhs: Expr<Rhs>) -> Self::Output {
                $Op::<$LhsType, $RhsType>::new_unchecked(self.to_expr().canonical(), rhs.canonical()).to_expr()
            }
        }
    };
}

binary_op! {
    #[method = bitand, rhs_to_canonical_dyn = to_canonical_dyn, expr_enum_u = BitAndU, expr_enum_s = BitAndS]
    impl<L, R> BitAnd for _
    where
        L: IntTypeTrait,
        R: IntTypeTrait<Signed = L::Signed>,
}

binary_op! {
    #[method = bitor, rhs_to_canonical_dyn = to_canonical_dyn, expr_enum_u = BitOrU, expr_enum_s = BitOrS]
    impl<L, R> BitOr for _
    where
        L: IntTypeTrait,
        R: IntTypeTrait<Signed = L::Signed>,
}

binary_op! {
    #[method = bitxor, rhs_to_canonical_dyn = to_canonical_dyn, expr_enum_u = BitXorU, expr_enum_s = BitXorS]
    impl<L, R> BitXor for _
    where
        L: IntTypeTrait,
        R: IntTypeTrait<Signed = L::Signed>,
}

binary_op! {
    #[method = add, rhs_to_canonical_dyn = to_canonical_dyn, expr_enum_u = AddU, expr_enum_s = AddS]
    impl<L, R> Add for _
    where
        L: IntTypeTrait,
        R: IntTypeTrait<Signed = L::Signed>,
}

binary_op! {
    #[method = sub, rhs_to_canonical_dyn = to_canonical_dyn, expr_enum_u = SubU, expr_enum_s = SubS]
    impl<L, R> Sub for _
    where
        L: IntTypeTrait,
        R: IntTypeTrait<Signed = L::Signed>,
}

binary_op! {
    #[method = mul, rhs_to_canonical_dyn = to_canonical_dyn, expr_enum_u = MulU, expr_enum_s = MulS]
    impl<L, R> Mul for _
    where
        L: IntTypeTrait,
        R: IntTypeTrait<Signed = L::Signed>,
}

macro_rules! dyn_shift_op {
    (
        #[method = $Op:ident::$op:ident, expr_enum_u = $expr_enum_u:ident, expr_enum_s = $expr_enum_s:ident]
        impl $DynOp:ident
    ) => {
        fixed_ary_op! {
            pub struct $DynOp<LhsType,>
            where (
                LhsType: IntTypeTrait,
            )
            {
                pub lhs: Expr<LhsType::CanonicalValue>,
                pub rhs: Expr<DynUInt>,
                #[type]
                ty: <<Valueless<LhsType> as ops::$Op<Valueless<DynUIntType>>>::Output as ValuelessTr>::Type = ops::$Op::$op(
                    lhs.valueless(),
                    rhs.valueless(),
                ).ty,
                #[cache]
                literal_bits: Result<Interned<BitSlice>, ()> = {
                    lhs.to_literal_bits()
                        .ok()
                        .zip(rhs.to_literal_bits().ok())
                        .map(|(lhs, rhs)| {
                            ops::$Op::$op(
                                LhsType::CanonicalValue::from_bit_slice(&lhs),
                                DynUInt::from_bit_slice(&rhs),
                            )
                            .to_bits()
                        })
                        .ok_or(())
                },
                fn simulate(&self, sim_state: &mut SimState) -> _ {
                    ops::$Op::$op(self.lhs.simulate(sim_state), self.rhs.simulate(sim_state))
                }
                fn expr_enum(&self) -> _ {
                    struct Tag;
                    impl ConstBoolDispatchTag for Tag {
                        type Type<Select: GenericConstBool> = $DynOp<DynIntType<Select>>;
                    }
                    match ConstBoolDispatch::new::<Tag, LhsType::Signed>(self.canonical()) {
                        ConstBoolDispatch::False(v) => ExprEnum::$expr_enum_u(v.intern_sized()),
                        ConstBoolDispatch::True(v) => ExprEnum::$expr_enum_s(v.intern_sized()),
                    }
                }
                fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
                    self.literal_bits
                }
            }
        }

        impl<Lhs: Value<Type = LhsType>, LhsType, RhsType, Rhs: Value<Type = RhsType>> ops::$Op<Expr<Rhs>> for Expr<Lhs>
        where
            LhsType: IntTypeTrait,
            RhsType: IntTypeTrait<Signed = ConstBool<false>>,
        {
            type Output = Expr<<<Valueless<LhsType> as ops::$Op<Valueless<RhsType>>>::Output as ValuelessTr>::Value>;

            fn $op(self, rhs: Expr<Rhs>) -> Self::Output {
                $DynOp::<LhsType>::new_unchecked(self.canonical(), rhs.canonical()).to_expr()
            }
        }

        impl<LhsType, RhsType, Rhs: Value<Type = RhsType>> ops::$Op<Expr<Rhs>> for IntValue<LhsType>
        where
            LhsType: IntTypeTrait,
            RhsType: IntTypeTrait<Signed = ConstBool<false>>,
        {
            type Output = Expr<<<Valueless<LhsType> as ops::$Op<Valueless<RhsType>>>::Output as ValuelessTr>::Value>;

            fn $op(self, rhs: Expr<Rhs>) -> Self::Output {
                $DynOp::<LhsType>::new_unchecked(self.to_expr().canonical(), rhs.canonical()).to_expr()
            }
        }
    };
}

macro_rules! fixed_shift_op {
    (
        #[method = $Op:ident::$op:ident, expr_enum_u = $expr_enum_u:ident, expr_enum_s = $expr_enum_s:ident]
        impl $FixedOp:ident
    ) => {
        fixed_ary_op! {
            pub struct $FixedOp<LhsType,>
            where (
                LhsType: IntTypeTrait,
            )
            {
                pub lhs: Expr<LhsType::CanonicalValue>,
                pub rhs: usize,
                #[type]
                ty: <<Valueless<LhsType> as ops::$Op<usize>>::Output as ValuelessTr>::Type = ops::$Op::$op(
                    lhs.valueless(),
                    rhs,
                ).ty,
                #[cache]
                literal_bits: Result<Interned<BitSlice>, ()> = {
                    lhs.to_literal_bits()
                        .map(|lhs| {
                            ops::$Op::$op(
                                LhsType::CanonicalValue::from_bit_slice(&lhs),
                                rhs,
                            )
                            .to_bits()
                        })
                },
                fn simulate(&self, sim_state: &mut SimState) -> _ {
                    ops::$Op::$op(self.lhs.simulate(sim_state), self.rhs)
                }
                fn expr_enum(&self) -> _ {
                    struct Tag;
                    impl ConstBoolDispatchTag for Tag {
                        type Type<Select: GenericConstBool> = $FixedOp<DynIntType<Select>>;
                    }
                    match ConstBoolDispatch::new::<Tag, LhsType::Signed>(self.canonical()) {
                        ConstBoolDispatch::False(v) => ExprEnum::$expr_enum_u(v.intern_sized()),
                        ConstBoolDispatch::True(v) => ExprEnum::$expr_enum_s(v.intern_sized()),
                    }
                }
                fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
                    self.literal_bits
                }
            }
        }

        impl<Lhs: Value<Type = LhsType>, LhsType> ops::$Op<usize> for Expr<Lhs>
        where
            LhsType: IntTypeTrait,
        {
            type Output = Expr<<<Valueless<LhsType> as ops::$Op<usize>>::Output as ValuelessTr>::Value>;

            fn $op(self, rhs: usize) -> Self::Output {
                $FixedOp::<LhsType>::new_unchecked(self.canonical(), rhs).to_expr()
            }
        }
    };
}

dyn_shift_op! {
    #[method = Shl::shl, expr_enum_u = DynShlU, expr_enum_s = DynShlS]
    impl DynShl
}

dyn_shift_op! {
    #[method = Shr::shr, expr_enum_u = DynShrU, expr_enum_s = DynShrS]
    impl DynShr
}

fixed_shift_op! {
    #[method = Shl::shl, expr_enum_u = FixedShlU, expr_enum_s = FixedShlS]
    impl FixedShl
}

fixed_shift_op! {
    #[method = Shr::shr, expr_enum_u = FixedShrU, expr_enum_s = FixedShrS]
    impl FixedShr
}

macro_rules! cmp_op {
    ($($CmpOp:ident, $CmpOpU:ident, $CmpOpS:ident, $fn:ident, $Op:ident::$op:ident;)+) => {
        $(fixed_ary_op! {
            pub struct $CmpOp<LhsType, RhsType, Output,>
            where (
                LhsType: IntTypeTrait,
                RhsType: IntTypeTrait<Signed = LhsType::Signed>,
                Output: FixedOrDynIntType<1, Signed = ConstBool<false>>,
            )
            {
                pub lhs: Expr<LhsType::CanonicalValue>,
                pub rhs: Expr<RhsType::CanonicalValue>,
                #[type]
                ty: Output = FixedOrDynIntType::new(),
                #[cache]
                literal_bits: Result<Interned<BitSlice>, ()> = {
                    lhs.to_literal_bits()
                        .ok()
                        .zip(rhs.to_literal_bits().ok())
                        .map(|(lhs, rhs)| {
                            interned_bit(
                                LhsType::CanonicalValue::from_bit_slice(&lhs)
                                    .value()
                                    .$op(&RhsType::CanonicalValue::from_bit_slice(&rhs).value()),
                            )
                        })
                        .ok_or(())
                },
                fn simulate(&self, sim_state: &mut SimState) -> _ {
                    self.lhs.simulate(sim_state).$fn(self.rhs.simulate(sim_state)).into_canonical()
                }

                fn expr_enum(&self) -> _ {
                    struct Tag;
                    impl ConstBoolDispatchTag for Tag {
                        type Type<Select: GenericConstBool> = $CmpOp<DynIntType<Select>, DynIntType<Select>, DynUIntType>;
                    }
                    match ConstBoolDispatch::new::<Tag, LhsType::Signed>(self.canonical()) {
                        ConstBoolDispatch::False(v) => ExprEnum::$CmpOpU(v.intern_sized()),
                        ConstBoolDispatch::True(v) => ExprEnum::$CmpOpS(v.intern_sized()),
                    }
                }

                fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
                    self.literal_bits
                }
            }
        })*

        impl<LhsType: IntTypeTrait, RhsType: IntTypeTrait<Signed = LhsType::Signed>, Rhs: Value<Type = RhsType>, Lhs: Value<Type = LhsType>> IntCmp<Expr<Rhs>> for Expr<Lhs> {
            type Output = Expr<UInt<1>>;

            $(fn $fn(self, rhs: Expr<Rhs>) -> Self::Output {
                $CmpOp::<LhsType, RhsType, UIntType<1>>::new_unchecked(self.canonical(), rhs.canonical()).to_expr()
            })*
        }

        impl<LhsType: IntTypeTrait, RhsType: IntTypeTrait<Signed = LhsType::Signed>, Rhs: Value<Type = RhsType>, Lhs: Value<Type = LhsType>> IntCmp<Rhs> for Expr<Lhs> {
            type Output = Expr<UInt<1>>;

            $(fn $fn(self, rhs: Rhs) -> Self::Output {
                self.$fn(rhs.to_expr())
            })*
        }

        impl<LhsType: IntTypeTrait, RhsType: IntTypeTrait<Signed = LhsType::Signed>, Rhs: Value<Type = RhsType>> IntCmp<Expr<Rhs>> for IntValue<LhsType> {
            type Output = Expr<UInt<1>>;

            $(fn $fn(self, rhs: Expr<Rhs>) -> Self::Output {
                self.to_expr().$fn(rhs)
            })*
        }

        impl<LhsType: IntTypeTrait, RhsType: IntTypeTrait<Signed = LhsType::Signed>> IntCmp<Valueless<RhsType>> for Valueless<LhsType> {
            type Output = Valueless<UIntType<1>>;

            $(fn $fn(self, _rhs: Valueless<RhsType>) -> Self::Output {
                Valueless { ty: UIntType::new() }
            })*
        }
    };
}

cmp_op! {
    CmpLt, CmpLtU, CmpLtS, cmp_lt, PartialOrd::lt;
    CmpLe, CmpLeU, CmpLeS, cmp_le, PartialOrd::le;
    CmpGt, CmpGtU, CmpGtS, cmp_gt, PartialOrd::gt;
    CmpGe, CmpGeU, CmpGeS, cmp_ge, PartialOrd::ge;
    CmpEq, CmpEqU, CmpEqS, cmp_eq, PartialEq::eq;
    CmpNe, CmpNeU, CmpNeS, cmp_ne, PartialEq::ne;
}

fixed_ary_op! {
    pub struct CastInt<FromType, ToType,>
    where (
        FromType: IntTypeTrait,
        ToType: IntTypeTrait,
    )
    {
        pub value: Expr<FromType::CanonicalValue>,
        #[type(ty)]
        pub ty: ToType = ty,
        #[cache]
        literal_bits: Result<Interned<BitSlice>, ()> = {
            value.to_literal_bits().map(|literal_bits| {
                let mut bits = literal_bits.to_bitvec();
                let fill = FromType::Signed::VALUE && bits.len().checked_sub(1).map(|i| bits[i]).unwrap_or(false);
                bits.resize(ty.width(), fill);
                Intern::intern_owned(bits)
            })
        },

        fn simulate(&self, sim_state: &mut SimState) -> _ {
            self.value.simulate(sim_state).cast(self.ty.canonical())
        }

        fn expr_enum(&self) -> _ {
            struct Tag1;
            impl ConstBoolDispatchTag for Tag1 {
                type Type<Select: GenericConstBool> = ConstBoolDispatch<
                    CastInt<DynIntType<Select>, DynUIntType>,
                    CastInt<DynIntType<Select>, DynSIntType>,
                >;
            }
            struct Tag2<FromSigned: GenericConstBool>(FromSigned);
            impl<FromSigned: GenericConstBool> ConstBoolDispatchTag for Tag2<FromSigned> {
                type Type<Select: GenericConstBool> = CastInt<DynIntType<FromSigned>, DynIntType<Select>>;
            }
            match ConstBoolDispatch::new::<Tag1, FromType::Signed>(ConstBoolDispatch::new::<Tag2<FromType::Signed>, ToType::Signed>(self.canonical())) {
                ConstBoolDispatch::False(ConstBoolDispatch::False(v)) => ExprEnum::CastUIntToUInt(v.intern_sized()),
                ConstBoolDispatch::False(ConstBoolDispatch::True(v)) => ExprEnum::CastUIntToSInt(v.intern_sized()),
                ConstBoolDispatch::True(ConstBoolDispatch::False(v)) => ExprEnum::CastSIntToUInt(v.intern_sized()),
                ConstBoolDispatch::True(ConstBoolDispatch::True(v)) => ExprEnum::CastSIntToSInt(v.intern_sized()),
            }
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.literal_bits
        }
    }
}

impl<FromType: IntTypeTrait> Expr<IntValue<FromType>> {
    pub fn cast<
        ToType: IntTypeTrait<CanonicalType = DynIntType<<ToType as IntTypeTrait>::Signed>>,
    >(
        self,
        ty: ToType,
    ) -> Expr<ToType::Value> {
        CastInt::<FromType, ToType>::new_unchecked(self.canonical(), ty).to_expr()
    }
    pub fn as_same_width_uint(self) -> Expr<IntValue<FromType::SameWidthUInt>> {
        self.cast(self.ty().as_same_width_uint())
    }
    pub fn as_same_width_sint(self) -> Expr<IntValue<FromType::SameWidthSInt>> {
        self.cast(self.ty().as_same_width_sint())
    }
    pub fn as_same_value_uint(self) -> Expr<DynUInt> {
        self.cast(self.ty().as_same_value_uint())
    }
    pub fn as_same_value_sint(self) -> Expr<DynSInt> {
        self.cast(self.ty().as_same_value_sint())
    }
}

fixed_ary_op! {
    pub struct Slice<Base,>
    where (
        Base: IntTypeTrait,
    )
    {
        pub base: Expr<Base::CanonicalValue>,
        pub range: Range<usize>,
        #[type]
        ty: DynUIntType = base.valueless().slice(range.clone()).ty,
        #[cache]
        literal_bits: Result<Interned<BitSlice>, ()> = {
            base.to_literal_bits().map(|base| {
                base[range.clone()].intern()
            })
        },
        fn simulate(&self, sim_state: &mut SimState) -> _ {
            self.base.simulate(sim_state).slice(self.range.clone())
        }

        fn expr_enum(&self) -> _ {
            struct Tag;
            impl ConstBoolDispatchTag for Tag {
                type Type<Select: GenericConstBool> = Slice<DynIntType<Select>>;
            }
            match ConstBoolDispatch::new::<Tag, Base::Signed>(self.canonical()) {
                ConstBoolDispatch::False(v) => ExprEnum::SliceUInt(v.intern_sized()),
                ConstBoolDispatch::True(v) => ExprEnum::SliceSInt(v.intern_sized()),
            }
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.literal_bits
        }
    }
}

impl<T: IntTypeTrait> Expr<IntValue<T>> {
    pub fn slice<I: RangeBounds<usize>>(self, index: I) -> Expr<DynUInt> {
        Slice::<T>::new_unchecked(
            self.canonical(),
            self.canonical_type().slice_index_to_range(index),
        )
        .to_expr()
    }
    pub fn bit(self, index: usize) -> Expr<UInt<1>> {
        let width = self.canonical_type().width();
        if width == 0 {
            return false.to_expr();
        }
        let index = index.min(width - 1);
        self.slice(index..=index).cast(UIntType::new())
    }
    pub fn is_negative(self) -> Expr<UInt<1>> {
        if T::Signed::VALUE {
            false.to_expr()
        } else {
            self.bit(usize::MAX)
        }
    }
    pub fn zero() -> Self
    where
        T: Default,
    {
        IntValue::<T>::zero().to_expr()
    }
    pub fn is_zero(self) -> Expr<UInt<1>> {
        Expr::<Int<T::Signed, 1>>::zero().cmp_eq(self)
    }
    pub fn is_positive(self) -> Expr<UInt<1>> {
        Expr::<Int<T::Signed, 1>>::zero().cmp_lt(self)
    }
}

pub trait ExprIndex<I>: Value {
    type OutputValue: Value;

    fn expr_index(this: Expr<Self>, index: I) -> Expr<Self::OutputValue>;
}

impl<T: ExprIndex<I>, I> Index<I> for Expr<T> {
    type Output = Expr<T::OutputValue>;

    fn index(&self, index: I) -> &Self::Output {
        Interned::<_>::into_inner(T::expr_index(*self, index).intern_sized())
    }
}

mod range_sealed {
    pub trait Sealed {}
}

pub trait BuiltInRange<T: ?Sized>: RangeBounds<T> + range_sealed::Sealed {}

impl<I: ?Sized, T: RangeBounds<I> + ?Sized + range_sealed::Sealed> BuiltInRange<I> for T {}

macro_rules! impl_built_in_range {
    (<$T:ident> $ty:ty) => {
        impl<$T> range_sealed::Sealed for $ty {}
    };
    ($ty:ty) => {
        impl range_sealed::Sealed for $ty {}
    };
}

impl_built_in_range!(<T> (std::ops::Bound<T>, std::ops::Bound<T>));
impl_built_in_range!(<T> std::ops::Range<T>);
impl_built_in_range!(<T> std::ops::RangeFrom<T>);
impl_built_in_range!(std::ops::RangeFull);
impl_built_in_range!(<T> std::ops::RangeInclusive<T>);
impl_built_in_range!(<T> std::ops::RangeTo<T>);
impl_built_in_range!(<T> std::ops::RangeToInclusive<T>);

impl<T: IntTypeTrait, I: BuiltInRange<usize>> ExprIndex<I> for IntValue<T> {
    type OutputValue = DynUInt;

    fn expr_index(this: Expr<Self>, index: I) -> Expr<Self::OutputValue> {
        this.slice(index)
    }
}

impl<T: IntTypeTrait> ExprIndex<usize> for IntValue<T> {
    type OutputValue = UInt<1>;

    fn expr_index(this: Expr<Self>, index: usize) -> Expr<Self::OutputValue> {
        this.bit(index)
    }
}

macro_rules! reduce_bit_op {
    ($name:ident; |$bits:ident| $literal_bits_bool:expr) => {
        fixed_ary_op! {
            pub struct $name<Output,>
            where (
                Output: FixedOrDynIntType<1, Signed = ConstBool<false>>,
            )
            {
                pub arg: Expr<DynUInt>,
                #[type]
                ty: Output = FixedOrDynIntType::new(),
                #[cache]
                literal_bits: Result<Interned<BitSlice>, ()> = {
                    arg.to_literal_bits().map(|$bits| interned_bit($literal_bits_bool))
                },
                fn simulate(&self, sim_state: &mut SimState) -> _ {
                    todo!()
                }

                fn expr_enum(&self) -> _ {
                    ExprEnum::$name($name::new_unchecked(self.arg).intern_sized())
                }

                fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
                    self.literal_bits
                }
            }
        }
    };
}

reduce_bit_op!(ReduceBitAnd; |bits| bits.all());
reduce_bit_op!(ReduceBitOr; |bits| bits.any());
reduce_bit_op!(ReduceBitXor; |bits| bits.count_ones() % 2 == 1);

impl<T: IntTypeTrait<Signed = ConstBool<false>>> Expr<IntValue<T>> {
    pub fn reduce_bitand(self) -> Expr<UInt<1>> {
        ReduceBitAnd::<UIntType<1>>::new_unchecked(self.canonical()).to_expr()
    }
    pub fn reduce_bitor(self) -> Expr<UInt<1>> {
        ReduceBitOr::<UIntType<1>>::new_unchecked(self.canonical()).to_expr()
    }
    pub fn reduce_bitxor(self) -> Expr<UInt<1>> {
        ReduceBitXor::<UIntType<1>>::new_unchecked(self.canonical()).to_expr()
    }
}

fixed_ary_op! {
    pub struct FieldAccess<Base, Field,>
    where (
        Base: BundleType,
        Base::Value: BundleValue,
        Field: Type,
    )
    {
        base: Expr<Base::CanonicalValue>,
        name: Interned<str>,
        #[cache]
        base_ty: DynBundleType = base.canonical_type(),
        #[cache]
        index: usize = {
            let Some(&index) = base_ty.name_indexes().get(&name) else {
                panic!("bundle field {name:?} not found");
            };
            index
        },
        #[cache]
        field: FieldType<Interned<dyn DynCanonicalType>> = base_ty.fields()[index],
        #[type]
        ty: Field = {
            let fields = base_ty.fields();
            let field_ty = fields[index].ty.downcast::<Field::CanonicalType>();
            Field::from_canonical_type(field_ty.expect("field type doesn't match type generic"))
        },
        #[cache]
        literal_bits: Result<Interned<BitSlice>, ()> = {
            base.to_literal_bits().map(|base| {
                base[base_ty.field_offsets()[index]..][..field.ty.bit_width()].intern()
            })
        },
        #[target]
        target: Option<Interned<Target>> = base.target().map(|base| {
            Intern::intern_sized(base.join(TargetPathElement::intern_sized(TargetPathBundleField{name:field.name}.into())))
        }),
        fn simulate(&self, sim_state: &mut SimState) -> _ {
            todo!()
        }

        fn expr_enum(&self) -> _ {
            ExprEnum::FieldAccess(FieldAccess::new_unchecked(self.base, self.name).intern_sized())
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.literal_bits
        }
    }
}

impl<Base, Field> FieldAccess<Base, Field>
where
    Base: BundleType,
    Base::Value: BundleValue,
    Field: Type,
{
    pub fn base(&self) -> Expr<Base::CanonicalValue> {
        self.base
    }
    pub fn name(&self) -> Interned<str> {
        self.name
    }
    pub fn index(&self) -> usize {
        self.index
    }
}

fixed_ary_op! {
    pub struct VariantAccess<T, VariantTy,>
    where (
        T: EnumType,
        T::Value: EnumValue,
        VariantTy: Type,
    )
    {
        base: Expr<T::CanonicalValue>,
        variant_index: usize,
        #[cache]
        base_ty: DynEnumType = base.canonical_type(),
        #[type]
        ty: VariantTy = VariantTy::from_dyn_canonical_type(base_ty.variants()[variant_index].ty.unwrap_or_else(
            || ().canonical_dyn(),
        )),
        #[cache]
        literal_bits: Result<Interned<BitSlice>, ()> = {
            base.to_literal_bits().map(|base| {
                base[base_ty.discriminant_bit_width()..][..ty.bit_width()].intern()
            })
        },
        fn simulate(&self, sim_state: &mut SimState) -> _ {
            todo!()
        }

        fn expr_enum(&self) -> _ {
            ExprEnum::VariantAccess(VariantAccess::new_unchecked(
                self.base,
                self.variant_index,
            ).intern_sized())
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.literal_bits
        }
    }
}

impl<T, VariantTy> VariantAccess<T, VariantTy>
where
    T: EnumType,
    T::Value: EnumValue,
    VariantTy: Type,
{
    pub fn base(&self) -> Expr<T::CanonicalValue> {
        self.base
    }
    pub fn variant_index(&self) -> usize {
        self.variant_index
    }
    pub fn variant_type(&self) -> VariantType<Interned<dyn DynCanonicalType>> {
        self.base_ty.variants()[self.variant_index]
    }
}

impl<T> VariantAccess<T, Interned<dyn DynCanonicalType>>
where
    T: EnumType,
    T::Value: EnumValue,
{
    pub fn downcast_unchecked<VariantTy: Type>(&self) -> VariantAccess<T, VariantTy> {
        VariantAccess::new_unchecked(self.base, self.variant_index)
    }
}

fixed_ary_op! {
    pub struct CastToBits<> where () {
        value: Expr<DynCanonicalValue>,
        #[type]
        ty: DynUIntType = DynUIntType::new(value.ty().bit_width()),
        fn simulate(&self, sim_state: &mut SimState) -> _ {
            todo!()
        }

        fn expr_enum(&self) -> _ {
            ExprEnum::CastToBits(self.intern())
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.value.to_literal_bits()
        }
    }
}

impl CastToBits {
    pub fn value(&self) -> Expr<DynCanonicalValue> {
        self.value
    }
}

impl<V: Value> Expr<V> {
    pub fn cast_to_bits(self) -> Expr<DynUInt> {
        CastToBits::new_unchecked(self.to_canonical_dyn()).to_expr()
    }
}

fixed_ary_op! {
    pub struct CastBitsTo<T,> where (
        T: Type,
        T::Value: Value<Type = T>,
    ) {
        value: Expr<DynUInt>,
        #[type(ty)]
        ty: T = {
            assert!(ty.is_castable_from_bits(), "can't cast bits to type: {ty:?}");
            assert_eq!(value.ty().width, ty.bit_width(), "input bit width didn't match target type's bit width");
            ty
        },
        fn simulate(&self, sim_state: &mut SimState) -> _ {
            todo!()
        }

        fn expr_enum(&self) -> _ {
            ExprEnum::CastBitsTo(CastBitsTo::new_unchecked(self.value, self.ty.canonical_dyn()).intern_sized())
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.value.to_literal_bits()
        }
    }
}

impl<T> CastBitsTo<T>
where
    T: Type,
    T::Value: Value<Type = T>,
{
    pub fn value(&self) -> Expr<DynUInt> {
        self.value
    }
}

impl<T: IntTypeTrait<Signed = ConstBool<false>>> Expr<IntValue<T>> {
    pub fn cast_bits_to<U: Value>(self, ty: U::Type) -> Expr<U>
    where
        U::Type: Type<Value = U>,
    {
        CastBitsTo::new_unchecked(self.canonical(), ty).to_expr()
    }
}

macro_rules! cast_bit_to_typed_bit {
    (pub struct $Op:ident$(<$T:ident: $Bound:ident<$BoundArg:tt>>)?; $FromType:ty => $ToType:ty) => {
        fixed_ary_op! {
            pub struct $Op<$($T,)?>
            where (
                $($T: $Bound<$BoundArg>,)?
            )
            {
                pub value: Expr<<$FromType as Type>::Value>,
                #[type]
                pub ty: $ToType = <$ToType>::new(),
                fn simulate(&self, sim_state: &mut SimState) -> _ {
                    todo!()
                }

                fn expr_enum(&self) -> _ {
                    ExprEnum::$Op($Op::new_unchecked(self.value).intern_sized())
                }

                fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
                    self.value.to_literal_bits()
                }
            }
        }
    };
}

cast_bit_to_typed_bit!(pub struct CastBitToClock; UIntType<1> => ClockType);

impl ToClock for Expr<UInt<1>> {
    fn to_clock(&self) -> Expr<Clock> {
        CastBitToClock::new_unchecked(*self).to_expr()
    }
}

impl<T: Type> Valueless<T>
where
    Expr<T::Value>: ToClock,
{
    pub fn to_clock(self) -> Valueless<ClockType> {
        Valueless { ty: ClockType }
    }
}

cast_bit_to_typed_bit!(pub struct CastBitToSyncReset; UIntType<1> => SyncResetType);

impl ToSyncReset for Expr<UInt<1>> {
    fn to_sync_reset(&self) -> Expr<SyncReset> {
        CastBitToSyncReset::new_unchecked(*self).to_expr()
    }
}

impl<T: Type> Valueless<T>
where
    Expr<T::Value>: ToSyncReset,
{
    pub fn to_sync_reset(self) -> Valueless<SyncResetType> {
        Valueless { ty: SyncResetType }
    }
}

cast_bit_to_typed_bit!(pub struct CastBitToAsyncReset; UIntType<1> => AsyncResetType);

impl ToAsyncReset for Expr<UInt<1>> {
    fn to_async_reset(&self) -> Expr<AsyncReset> {
        CastBitToAsyncReset::new_unchecked(*self).to_expr()
    }
}

impl<T: Type> Valueless<T>
where
    Expr<T::Value>: ToAsyncReset,
{
    pub fn to_async_reset(self) -> Valueless<AsyncResetType> {
        Valueless { ty: AsyncResetType }
    }
}

cast_bit_to_typed_bit!(pub struct CastSyncResetToReset; SyncResetType => ResetType);

impl ToReset for Expr<SyncReset> {
    fn to_reset(&self) -> Expr<Reset> {
        CastSyncResetToReset::new_unchecked(*self).to_expr()
    }
}

cast_bit_to_typed_bit!(pub struct CastAsyncResetToReset; AsyncResetType => ResetType);

impl ToReset for Expr<AsyncReset> {
    fn to_reset(&self) -> Expr<Reset> {
        CastAsyncResetToReset::new_unchecked(*self).to_expr()
    }
}

impl<T: Type> Valueless<T>
where
    Expr<T::Value>: ToReset,
{
    pub fn to_reset(self) -> Valueless<ResetType> {
        Valueless { ty: ResetType }
    }
}

pub trait ToBit: CanonicalValue
where
    Self::Type: CanonicalType<CanonicalType = Self::Type, Value = Self, CanonicalValue = Self>,
{
    fn to_bit(v: Expr<Self>) -> Expr<UInt<1>>;
}

impl<V: ToBit> Expr<V>
where
    V::Type: CanonicalType<CanonicalType = V::Type, Value = V, CanonicalValue = V>,
{
    pub fn to_bit(self) -> Expr<UInt<1>> {
        V::to_bit(self)
    }
}

impl<V: ToBit> Valueless<V>
where
    V::Type: CanonicalType<CanonicalType = V::Type, Value = V, CanonicalValue = V>,
{
    pub fn to_bit(self) -> Valueless<UIntType<1>> {
        Valueless {
            ty: UIntType::new(),
        }
    }
}

cast_bit_to_typed_bit!(pub struct CastClockToBit<ToType: FixedOrDynIntType<1>>; ClockType => ToType);

impl ToBit for Clock {
    fn to_bit(v: Expr<Self>) -> Expr<UInt<1>> {
        CastClockToBit::<UIntType<1>>::new_unchecked(v).to_expr()
    }
}

cast_bit_to_typed_bit!(pub struct CastSyncResetToBit<ToType: FixedOrDynIntType<1>>; SyncResetType => ToType);

impl ToBit for SyncReset {
    fn to_bit(v: Expr<Self>) -> Expr<UInt<1>> {
        CastSyncResetToBit::<UIntType<1>>::new_unchecked(v).to_expr()
    }
}

cast_bit_to_typed_bit!(pub struct CastAsyncResetToBit<ToType: FixedOrDynIntType<1>>; AsyncResetType => ToType);

impl ToBit for AsyncReset {
    fn to_bit(v: Expr<Self>) -> Expr<UInt<1>> {
        CastAsyncResetToBit::<UIntType<1>>::new_unchecked(v).to_expr()
    }
}

cast_bit_to_typed_bit!(pub struct CastResetToBit<ToType: FixedOrDynIntType<1>>; ResetType => ToType);

impl ToBit for Reset {
    fn to_bit(v: Expr<Self>) -> Expr<UInt<1>> {
        CastResetToBit::<UIntType<1>>::new_unchecked(v).to_expr()
    }
}

fixed_ary_op! {
    pub struct ArrayLiteral<ArrayTy,>
    where (
        ArrayTy: ArrayTypeTrait,
    )
    {
        elements: Interned<[Expr<DynCanonicalValue>]>,
        #[type(ty)]
        ty: ArrayTy = {
            let canonical_element_ty = ty.borrow().element().canonical_dyn();
            assert_eq!(ty.borrow().len(), elements.len());
            for element in elements {
                assert_eq!(element.ty(), canonical_element_ty);
            }
            ty
        },
        #[cache]
        literal_bits: Result<Interned<BitSlice>, ()> = {
            let mut bits = Some(BitVec::with_capacity(ty.bit_width()));
            for element in elements {
                let Ok(element_bits) = element.to_literal_bits() else {
                    bits = None;
                    break;
                };
                bits.as_mut().unwrap().extend_from_bitslice(&element_bits);
            }
            bits.map(Intern::intern_owned).ok_or(())
        },
        fn simulate(&self, sim_state: &mut SimState) -> _ {
            todo!()
        }

        fn expr_enum(&self) -> _ {
            ExprEnum::ArrayLiteral(self.canonical().intern_sized())
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.literal_bits
        }
    }
}

impl<ArrayTy: ArrayTypeTrait<ValueArrayOrSlice = VA>, VA: ?Sized + ValueArrayOrSlice>
    ArrayLiteral<ArrayTy>
{
    pub fn element_ty(&self) -> &VA::ElementType {
        self.ty.borrow().element()
    }
    pub fn elements(&self) -> Interned<[Expr<DynCanonicalValue>]> {
        self.elements
    }
}

fixed_ary_op! {
    pub struct ArrayIndex<ElementType,>
    where (
        ElementType: Type,
    )
    {
        base: Expr<Array<[DynCanonicalValue]>>,
        index: usize,
        #[type]
        ty: ElementType = {
            let base_ty = ArrayType::<[ElementType::Value]>::from_canonical_type(base.ty());
            assert!(index < base_ty.len());
            base_ty.element().clone()
        },
        #[cache]
        literal_bits: Result<Interned<BitSlice>, ()> = {
            let base_ty = base.ty();
            let element_bit_width = base_ty.element().bit_width();
            base.to_literal_bits().map(|base| {
                let start = index * element_bit_width;
                let end = start + element_bit_width;
                base[start..end].intern()
            })
        },
        #[target]
        target: Option<Interned<Target>> = base.target().map(|base| {
            Intern::intern_sized(base.join(TargetPathElement::ArrayElement(TargetPathArrayElement { index }).intern()))
        }),
        fn simulate(&self, sim_state: &mut SimState) -> _ {
            todo!()
        }

        fn expr_enum(&self) -> _ {
            ExprEnum::ArrayIndex(ArrayIndex::new_unchecked(self.base, self.index).intern_sized())
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.literal_bits
        }
    }
}

impl<ElementType: Type> ArrayIndex<ElementType> {
    pub fn base(&self) -> Expr<Array<[DynCanonicalValue]>> {
        self.base
    }
    pub fn index(&self) -> usize {
        self.index
    }
}

fixed_ary_op! {
    pub struct DynArrayIndex<ElementType,>
    where (
        ElementType: Type,
    )
    {
        base: Expr<Array<[DynCanonicalValue]>>,
        index: Expr<DynUInt>,
        #[type]
        ty: ElementType = {
            let base_ty = ArrayType::<[ElementType::Value]>::from_canonical_type(base.ty());
            base_ty.element().clone()
        },
        #[cache]
        literal_bits: Result<Interned<BitSlice>, ()> = {
            let base_ty = base.ty();
            let element_bit_width = base_ty.element().bit_width();
            base.to_literal_bits().ok().zip(index.to_literal_bits().ok()).and_then(|(base, index)| {
                let start = DynUInt::from_bit_slice(&index).uint_value().to_usize()?.checked_mul(element_bit_width)?;
                let end = start.checked_add(element_bit_width)?;
                Some(base.get(start..end)?.intern())
            }).ok_or(())
        },
        #[target]
        target: Option<Interned<Target>> = base.target().map(|base| {
            Intern::intern_sized(base.join(TargetPathElement::DynArrayElement(TargetPathDynArrayElement {}).intern()))
        }),
        fn simulate(&self, sim_state: &mut SimState) -> _ {
            todo!()
        }

        fn expr_enum(&self) -> _ {
            ExprEnum::DynArrayIndex(DynArrayIndex::new_unchecked(self.base, self.index).intern_sized())
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.literal_bits
        }
    }
}

impl<ElementType: Type> DynArrayIndex<ElementType> {
    pub fn base(&self) -> Expr<Array<[DynCanonicalValue]>> {
        self.base
    }
    pub fn index(&self) -> Expr<DynUInt> {
        self.index
    }
}

impl<
        VA: ?Sized + ValueArrayOrSlice,
        Idx: IntTypeTrait<Signed = ConstBool<false>, CanonicalType = DynUIntType>,
    > ExprIndex<IntValue<Idx>> for Array<VA>
{
    type OutputValue = VA::Element;

    fn expr_index(this: Expr<Self>, index: IntValue<Idx>) -> Expr<Self::OutputValue> {
        let index = index.uint_value().try_into().expect("index out of range");
        ArrayIndex::<VA::ElementType>::new_unchecked(this.canonical(), index).to_expr()
    }
}

impl<
        VA: ?Sized + ValueArrayOrSlice,
        Idx: IntTypeTrait<Signed = ConstBool<false>, CanonicalType = DynUIntType>,
    > ExprIndex<Expr<IntValue<Idx>>> for Array<VA>
{
    type OutputValue = VA::Element;

    fn expr_index(this: Expr<Self>, index: Expr<IntValue<Idx>>) -> Expr<Self::OutputValue> {
        let index: Expr<DynUInt> = index.canonical();
        if let Ok(index) = index.to_literal_bits() {
            let index = DynUInt::from_bit_slice(&index)
                .uint_value()
                .try_into()
                .expect("index out of range");
            return ArrayIndex::<VA::ElementType>::new_unchecked(this.canonical(), index).to_expr();
        }
        DynArrayIndex::<VA::ElementType>::new_unchecked(this.canonical(), index).to_expr()
    }
}

impl<VA: ?Sized + ValueArrayOrSlice> ExprIndex<usize> for Array<VA> {
    type OutputValue = VA::Element;

    fn expr_index(this: Expr<Self>, index: usize) -> Expr<Self::OutputValue> {
        ArrayIndex::<VA::ElementType>::new_unchecked(this.canonical(), index).to_expr()
    }
}

fixed_ary_op! {
    pub struct BundleLiteral<BundleTy,>
    where (
        BundleTy: BundleType,
        BundleTy::Value: BundleValue<Type = BundleTy>,
    )
    {
        fields: Interned<[Expr<DynCanonicalValue>]>,
        #[type(ty)]
        ty: BundleTy = {
            assert!(
                ty.is_passive(),
                concat!(
                    "bundle literal must be a passive type (no ",
                    "#[flip] fields and all fields must be passive types)",
                ),
            );
            assert_eq!(
                ty.fields().len(),
                fields.len(),
                "field values don't match type"
            );
            for (field_ty, field) in ty.fields().iter().zip(fields.iter()) {
                assert_eq!(field_ty.ty, field.ty(), "field value doesn't match type");
            }
            ty
        },
        #[cache]
        literal_bits: Result<Interned<BitSlice>, ()> = {
            let mut bits = Some(BitVec::new());
            for field in fields {
                let Ok(v) = field.to_literal_bits() else {
                    bits = None;
                    break;
                };
                bits.as_mut().unwrap().extend_from_bitslice(&v);
            }
            bits.map(Intern::intern_owned).ok_or(())
        },
        fn simulate(&self, sim_state: &mut SimState) -> _ {
            todo!()
        }

        fn expr_enum(&self) -> _ {
            ExprEnum::BundleLiteral(self.canonical().intern_sized())
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.literal_bits
        }
    }
}

impl<BundleTy> BundleLiteral<BundleTy>
where
    BundleTy: BundleType,
    BundleTy::Value: BundleValue<Type = BundleTy>,
{
    pub fn fields(&self) -> Interned<[Expr<DynCanonicalValue>]> {
        self.fields
    }
}

fixed_ary_op! {
    pub struct EnumLiteral<EnumTy,>
    where (
        EnumTy: EnumType,
        EnumTy::Value: EnumValue<Type = EnumTy>,
    )
    {
        variant_value: Option<Expr<DynCanonicalValue>>,
        variant_index: usize,
        #[cache]
        variant_type: VariantType<Interned<dyn DynCanonicalType>> = ty.variants()[variant_index],
        #[type(ty)]
        ty: EnumTy = {
            assert_eq!(
                variant_type.ty,
                variant_value.map(|v| v.ty()),
                "variant value doesn't match type"
            );
            ty
        },
        #[cache]
        literal_bits: Result<Interned<BitSlice>, ()> = ty.variant_to_bits(variant_index, variant_value.as_ref()),
        fn simulate(&self, sim_state: &mut SimState) -> _ {
            todo!()
        }

        fn expr_enum(&self) -> _ {
            ExprEnum::EnumLiteral(self.canonical().intern_sized())
        }

        fn to_literal_bits(&self) -> Result<Interned<BitSlice>, ()> {
            self.literal_bits
        }
    }
}

impl<EnumTy> EnumLiteral<EnumTy>
where
    EnumTy: EnumType,
    EnumTy::Value: EnumValue<Type = EnumTy>,
{
    pub fn variant_value(&self) -> Option<Expr<DynCanonicalValue>> {
        self.variant_value
    }
    pub fn variant_index(&self) -> usize {
        self.variant_index
    }
    pub fn variant_type(&self) -> VariantType<Interned<dyn DynCanonicalType>> {
        self.variant_type
    }
    pub fn variant_name(&self) -> Interned<str> {
        self.variant_type.name
    }
}
